"""Find my iPhone test constants."""

from typing import Any

from .const import CLIENT_ID
from .const_account_family import (
    FIRST_NAME,
    FULL_NAME,
    LAST_NAME,
    MEMBER_1_APPLE_ID,
    MEMBER_1_FIRST_NAME,
    MEMBER_1_LAST_NAME,
    MEMBER_1_PERSON_ID,
    MEMBER_2_APPLE_ID,
    MEMBER_2_FIRST_NAME,
    MEMBER_2_LAST_NAME,
    MEMBER_2_PERSON_ID,
    PERSON_ID,
)

# Fakers
UUID = "ABCDEFGH-1234-5678-1234-ABCDEFGHIJKL"
LOCATION_LATITUDE = 45.123456789012345
LOCATION_LONGITUDE = 6.1234567890123456

IPHONE4_1 = "iPhone4,1"
IPHONE12_1 = "iPhone12,1"
MACBOOKPRO10_1 = "MacBookPro10,1"
IPAD7_3 = "iPad7,3"
MACBOOKPRO15_1 = "MacBookPro15,1"
MACBOOK_PRO_15 = "MacBook Pro 15"
MACBOOK_PRO = "MacBook Pro"

# Data
# Re-generated device :
# id = rawDeviceModel + prsId (if not None)
# baUUID = UUID + id
# So they can still be faked and unique
FMI_FAMILY_WORKING: dict[str, Any] = {
    "userInfo": {
        "accountFormatter": 0,
        "firstName": FIRST_NAME,
        "lastName": LAST_NAME,
        "membersInfo": {
            MEMBER_1_PERSON_ID: {
                "accountFormatter": 0,
                "firstName": MEMBER_1_FIRST_NAME,
                "lastName": MEMBER_1_LAST_NAME,
                "deviceFetchStatus": "DONE",
                "useAuthWidget": True,
                "isHSA": True,
                "appleId": MEMBER_1_APPLE_ID,
            },
            MEMBER_2_PERSON_ID: {
                "accountFormatter": 0,
                "firstName": MEMBER_2_FIRST_NAME,
                "lastName": MEMBER_2_LAST_NAME,
                "deviceFetchStatus": "DONE",
                "useAuthWidget": True,
                "isHSA": True,
                "appleId": MEMBER_2_APPLE_ID,
            },
        },
        "hasMembers": True,
    },
    "serverContext": {
        "minCallbackIntervalInMS": 5000,
        "enable2FAFamilyActions": False,
        "preferredLanguage": "fr-fr",
        "lastSessionExtensionTime": None,
        "enableMapStats": True,
        "callbackIntervalInMS": 2000,
        "validRegion": True,
        "timezone": {
            "currentOffset": -25200000,
            "previousTransition": 1583661599999,
            "previousOffset": -28800000,
            "tzCurrentName": "Pacific Daylight Time",
            "tzName": "America/Los_Angeles",
        },
        "authToken": None,
        "maxCallbackIntervalInMS": 60000,
        "classicUser": False,
        "isHSA": True,
        "trackInfoCacheDurationInSecs": 86400,
        "imageBaseUrl": "https://statici.icloud.com",
        "minTrackLocThresholdInMts": 100,
        "maxLocatingTime": 90000,
        "sessionLifespan": 900000,
        "info": "info_id",
        "prefsUpdateTime": 1413548552466,
        "useAuthWidget": True,
        "clientId": CLIENT_ID,
        "enable2FAFamilyRemove": False,
        "serverTimestamp": 1585867038112,
        "deviceImageVersion": "4",
        "macCount": 0,
        "deviceLoadStatus": "200",
        "maxDeviceLoadTime": 60000,
        "prsId": PERSON_ID,
        "showSllNow": False,
        "cloudUser": True,
        "enable2FAErase": False,
    },
    "alert": None,
    "userPreferences": {
        "webPrefs": {
            "id": "web_prefs",
            "selectedDeviceId": IPHONE4_1,
        }
    },
    "content": [
        {
            "msg": {
                "strobe": False,
                "userText": False,
                "playSound": True,
                "vibrate": True,
                "createTimestamp": 1584520568680,
                "statusCode": "200",
            },
            "canWipeAfterLock": True,
            "baUUID": UUID + IPHONE12_1,
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": True,
            "passcodeLength": 6,
            "deviceStatus": "200",
            "deviceColor": "1-6-0",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": False,
                "LST": True,
                "LKM": False,
                "WMG": True,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": False,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": True,
            "rawDeviceModel": IPHONE12_1,
            "id": IPHONE12_1,
            "remoteLock": None,
            "isLocating": True,
            "modelDisplayName": "iPhone",
            "lostTimestamp": "",
            "batteryLevel": 0.8299999833106995,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": {"createTimestamp": 1584520568680, "statusCode": "200"},
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPhone 11",
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "NotCharging",
            "trackingInfo": None,
            "name": "iPhone de " + FIRST_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPhone",
            "location": {
                "isOld": False,
                "isInaccurate": False,
                "altitude": 0.0,
                "positionType": "GPS",
                "latitude": LOCATION_LATITUDE,
                "floorLevel": 0,
                "horizontalAccuracy": 4.5370291025030465,
                "locationType": "",
                "timeStamp": 1585867037749,
                "locationFinished": True,
                "verticalAccuracy": 0.0,
                "longitude": LOCATION_LONGITUDE,
            },
            "deviceModel": "iphone11-1-6-0",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432463,
                "statusCode": "205",
            },
            "canWipeAfterLock": True,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 4,
            "deviceStatus": "203",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": True,
                "LST": True,
                "LKM": False,
                "WMG": False,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": IPHONE4_1,
            "id": IPHONE4_1,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iPhone",
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": {"createTimestamp": 1583057432463, "statusCode": "205"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": {
                "stopLostMode": False,
                "emailUpdates": True,
                "userText": True,
                "sound": False,
                "ownerNbr": "",
                "text": "",
                "createTimestamp": 1463594549526,
                "statusCode": "2201",
            },
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPhone 4s",
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "iPhone " + FULL_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPhone",
            "location": None,
            "deviceModel": "FifthGen",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432463,
                "statusCode": "205",
            },
            "canWipeAfterLock": True,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 4,
            "deviceStatus": "203",
            "deviceColor": "white",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": True,
                "LST": True,
                "LKM": False,
                "WMG": False,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": "iPod4,1",
            "id": "iPod4,1",
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iPod",
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": {"createTimestamp": 1583057432463, "statusCode": "205"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPod touch (4th generation)",
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "iPod Touch 4 " + MEMBER_2_FIRST_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPod",
            "location": None,
            "deviceModel": "FourthGen-white",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": False,
                "playSound": True,
                "vibrate": False,
                "createTimestamp": 1398963329049,
                "statusCode": "200",
            },
            "canWipeAfterLock": False,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 4,
            "deviceStatus": "203",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": True,
                "LKL": True,
                "LST": False,
                "LKM": True,
                "WMG": False,
                "PSS": False,
                "PIN": True,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": True,
                "KPD": True,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": MACBOOKPRO10_1,
            "id": MACBOOKPRO10_1,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": MACBOOK_PRO,
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": {"createTimestamp": 1398963329049, "statusCode": "200"},
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": MACBOOK_PRO_15,
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "Retina " + MEMBER_2_FIRST_NAME,
            "isMac": True,
            "thisDevice": False,
            "deviceClass": "MacBookPro",
            "location": None,
            "deviceModel": "MacBookPro10_1",
            "maxMsgChar": 500,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432463,
                "statusCode": "200",
            },
            "canWipeAfterLock": False,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 6,
            "deviceStatus": "203",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": True,
                "LKL": True,
                "LST": False,
                "LKM": True,
                "WMG": False,
                "PSS": False,
                "PIN": True,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": True,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": "MacBookPro11,3",
            "id": "MacBookPro11,3",
            "remoteLock": {"createTimestamp": 1433338956786, "statusCode": "2201"},
            "isLocating": False,
            "modelDisplayName": MACBOOK_PRO,
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": {"createTimestamp": 1583057432463, "statusCode": "200"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": MACBOOK_PRO_15,
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "Retina " + FIRST_NAME,
            "isMac": True,
            "thisDevice": False,
            "deviceClass": "MacBookPro",
            "location": None,
            "deviceModel": "MacBookPro11_3",
            "maxMsgChar": 500,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432463,
                "statusCode": "200",
            },
            "canWipeAfterLock": False,
            "baUUID": UUID + MACBOOKPRO15_1,
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": True,
            "passcodeLength": 6,
            "deviceStatus": "201",
            "deviceColor": "spacegray",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": True,
                "LKL": False,
                "LST": False,
                "LKM": True,
                "WMG": False,
                "PSS": False,
                "PIN": True,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": True,
                "KPD": True,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": MACBOOKPRO15_1,
            "id": MACBOOKPRO15_1,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": MACBOOK_PRO,
            "lostTimestamp": "",
            "batteryLevel": 0.26968246698379517,
            "mesg": {"createTimestamp": 1583057432463, "statusCode": "200"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": MACBOOK_PRO_15,
            "prsId": None,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Charging",
            "trackingInfo": None,
            "name": "MacBook Pro de " + FIRST_NAME,
            "isMac": True,
            "thisDevice": False,
            "deviceClass": "MacBookPro",
            "location": {
                "isOld": False,
                "isInaccurate": False,
                "altitude": 0.0,
                "positionType": "Wifi",
                "latitude": LOCATION_LATITUDE,
                "floorLevel": 0,
                "horizontalAccuracy": 65.0,
                "locationType": "",
                "timeStamp": 1585867020040,
                "locationFinished": False,
                "verticalAccuracy": 0.0,
                "longitude": LOCATION_LONGITUDE,
            },
            "deviceModel": "MacBookPro15_1-spacegray",
            "maxMsgChar": 500,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": None,
            "canWipeAfterLock": False,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 6,
            "deviceStatus": "200",
            "deviceColor": "0",
            "features": {
                "BTR": True,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": False,
                "LST": False,
                "LKM": False,
                "WMG": False,
                "PSS": True,
                "PIN": False,
                "LCK": False,
                "REM": False,
                "MCS": True,
                "KEY": False,
                "KPD": False,
                "WIP": False,
            },
            "lowPowerMode": False,
            "rawDeviceModel": "AirPods_8207",
            "id": "AirPods_8207",
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "Accessory",
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": "Accessory",
            "prsId": None,
            "audioChannels": [
                {"name": "left", "available": 1, "playing": False, "muted": False},
                {"name": "right", "available": 1, "playing": False, "muted": False},
            ],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "AirPods de " + FULL_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "Accessory",
            "location": {
                "isOld": False,
                "isInaccurate": False,
                "altitude": 0.0,
                "positionType": "GPS",
                "latitude": LOCATION_LATITUDE,
                "floorLevel": 0,
                "horizontalAccuracy": 4.5370291025030465,
                "locationType": "",
                "timeStamp": 1585867037749,
                "locationFinished": True,
                "verticalAccuracy": 0.0,
                "longitude": LOCATION_LONGITUDE,
            },
            "deviceModel": "AirPods_8207-0",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": None,
            "canWipeAfterLock": False,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 4,
            "deviceStatus": "201",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": True,
                "LKL": False,
                "LST": False,
                "LKM": True,
                "WMG": False,
                "PSS": False,
                "PIN": True,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": True,
                "KPD": True,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": MACBOOKPRO10_1,
            "id": MACBOOKPRO10_1 + MEMBER_2_PERSON_ID,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": MACBOOK_PRO,
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": MACBOOK_PRO_15,
            "prsId": MEMBER_2_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "MacBook Pro de " + MEMBER_2_FIRST_NAME,
            "isMac": True,
            "thisDevice": False,
            "deviceClass": "MacBookPro",
            "location": None,
            "deviceModel": "MacBookPro10_1",
            "maxMsgChar": 500,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": None,
            "canWipeAfterLock": True,
            "baUUID": UUID + IPHONE12_1 + MEMBER_2_PERSON_ID,
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": True,
            "passcodeLength": 6,
            "deviceStatus": "200",
            "deviceColor": "1-7-0",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": False,
                "LST": True,
                "LKM": False,
                "WMG": True,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": IPHONE12_1,
            "id": IPHONE12_1 + MEMBER_2_PERSON_ID,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iPhone",
            "lostTimestamp": "",
            "batteryLevel": 0.3400000035762787,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": True,
            "snd": None,
            "fmlyShare": False,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPhone 11",
            "prsId": MEMBER_2_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "NotCharging",
            "trackingInfo": None,
            "name": "iPhone " + MEMBER_2_FIRST_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPhone",
            "location": None,
            "deviceModel": "iphone11-1-7-0",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432335,
                "statusCode": "200",
            },
            "canWipeAfterLock": False,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 6,
            "deviceStatus": "203",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": False,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": True,
                "LKL": True,
                "LST": False,
                "LKM": True,
                "WMG": False,
                "PSS": False,
                "PIN": True,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": True,
                "KPD": True,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": "iMac10,1",
            "id": "iMac10,1" + MEMBER_1_PERSON_ID,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iMac",
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": {"createTimestamp": 1583057432335, "statusCode": "200"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": True,
            "lostDevice": None,
            "lostModeCapable": False,
            "wipedTimestamp": None,
            "deviceDisplayName": "iMac",
            "prsId": MEMBER_1_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": 'iMac 27" ' + MEMBER_1_LAST_NAME,
            "isMac": True,
            "thisDevice": False,
            "deviceClass": "iMac",
            "location": None,
            "deviceModel": "iMac10_1",
            "maxMsgChar": 500,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": None,
            "canWipeAfterLock": True,
            "baUUID": UUID + IPAD7_3 + MEMBER_1_PERSON_ID,
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": True,
            "passcodeLength": 6,
            "deviceStatus": "201",
            "deviceColor": "2-2-0",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": True,
                "LST": True,
                "LKM": False,
                "WMG": True,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": IPAD7_3,
            "id": IPAD7_3 + MEMBER_1_PERSON_ID,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iPad",
            "lostTimestamp": "",
            "batteryLevel": 0.3799999952316284,
            "mesg": None,
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": True,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPad Pro",
            "prsId": MEMBER_1_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "NotCharging",
            "trackingInfo": None,
            "name": "iPad " + MEMBER_1_LAST_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPad",
            "location": None,
            "deviceModel": "NinthGen-2-2-0",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432335,
                "statusCode": "205",
            },
            "canWipeAfterLock": True,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": False,
            "passcodeLength": 4,
            "deviceStatus": "203",
            "deviceColor": None,
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": True,
                "LST": True,
                "LKM": False,
                "WMG": False,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": False,
            "rawDeviceModel": IPHONE4_1,
            "id": IPHONE4_1 + MEMBER_1_PERSON_ID,
            "remoteLock": None,
            "isLocating": False,
            "modelDisplayName": "iPhone",
            "lostTimestamp": "",
            "batteryLevel": 0.0,
            "mesg": {"createTimestamp": 1583057432335, "statusCode": "205"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": True,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPhone 4s",
            "prsId": MEMBER_1_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "Unknown",
            "trackingInfo": None,
            "name": "iPhone",
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPhone",
            "location": None,
            "deviceModel": "FifthGen",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
        {
            "msg": {
                "strobe": False,
                "userText": True,
                "playSound": False,
                "vibrate": False,
                "createTimestamp": 1583057432335,
                "statusCode": "200",
            },
            "canWipeAfterLock": True,
            "baUUID": "",
            "wipeInProgress": False,
            "lostModeEnabled": False,
            "activationLocked": True,
            "passcodeLength": 6,
            "deviceStatus": "201",
            "deviceColor": "e1e4e3-d7d9d8",
            "features": {
                "BTR": False,
                "LLC": False,
                "CLK": False,
                "TEU": True,
                "SND": True,
                "CLT": False,
                "SVP": False,
                "SPN": False,
                "XRM": False,
                "CWP": False,
                "MSG": True,
                "LOC": True,
                "LMG": False,
                "LKL": False,
                "LST": True,
                "LKM": False,
                "WMG": True,
                "PSS": False,
                "PIN": False,
                "LCK": True,
                "REM": True,
                "MCS": False,
                "KEY": False,
                "KPD": False,
                "WIP": True,
            },
            "lowPowerMode": True,
            "rawDeviceModel": "iPhone6,2",
            "id": "iPhone6,2" + MEMBER_1_PERSON_ID,
            "remoteLock": None,
            "isLocating": True,
            "modelDisplayName": "iPhone",
            "lostTimestamp": "",
            "batteryLevel": 0.800000011920929,
            "mesg": {"createTimestamp": 1583057432335, "statusCode": "200"},
            "locationEnabled": True,
            "lockedTimestamp": None,
            "locFoundEnabled": False,
            "snd": None,
            "fmlyShare": True,
            "lostDevice": None,
            "lostModeCapable": True,
            "wipedTimestamp": None,
            "deviceDisplayName": "iPhone 5s",
            "prsId": MEMBER_1_PERSON_ID,
            "audioChannels": [],
            "locationCapable": True,
            "batteryStatus": "NotCharging",
            "trackingInfo": None,
            "name": "iPhone de " + MEMBER_1_FIRST_NAME,
            "isMac": False,
            "thisDevice": False,
            "deviceClass": "iPhone",
            "location": {
                "isOld": False,
                "isInaccurate": False,
                "altitude": 0.0,
                "positionType": "GPS",
                "latitude": LOCATION_LATITUDE,
                "floorLevel": 0,
                "horizontalAccuracy": 50.0,
                "locationType": "",
                "timeStamp": 1585866941186,
                "locationFinished": False,
                "verticalAccuracy": 0.0,
                "longitude": LOCATION_LONGITUDE,
            },
            "deviceModel": "5s-e1e4e3-d7d9d8",
            "maxMsgChar": 160,
            "darkWake": False,
            "remoteWipe": None,
        },
    ],
    "statusCode": "200",
}



"""Cmdline tests."""
# pylint: disable=protected-access

import argparse
import pickle
from io import BytesIO
from pprint import pformat
from unittest.mock import MagicMock, PropertyMock, mock_open, patch

import pytest

from pyicloud.cmdline import (
    _create_parser,
    _display_device_message_option,
    _display_device_silent_message_option,
    _enable_lost_mode_option,
    _handle_2fa,
    _handle_2sa,
    _list_devices_option,
    _play_device_sound_option,
    create_pickled_data,
    main,
)
from pyicloud.services.findmyiphone import AppleDevice
from tests import PyiCloudSessionMock
from tests.const import (
    AUTHENTICATED_USER,
    FMI_FAMILY_WORKING,
    REQUIRES_2FA_USER,
    VALID_2FA_CODE,
    VALID_PASSWORD,
)


def test_no_arg() -> None:
    """Test no args."""
    with pytest.raises(SystemExit, match="2"):
        main()


def test_username_password_invalid() -> None:
    """Test username and password commands."""
    # No password supplied
    with (
        patch("getpass.getpass", return_value=None),
        patch("argparse.ArgumentParser.parse_args") as mock_parse_args,
        patch("builtins.open", new_callable=mock_open),
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
        pytest.raises(SystemExit, match="2"),
    ):
        mock_parse_args.return_value = argparse.Namespace(
            username="valid_user",
            password=None,
            debug=False,
            interactive=True,
            china_mainland=False,
            delete_from_keyring=False,
            loglevel="info",
            no_verify_ssl=False,
            http_proxy=None,
            https_proxy=None,
            session_dir="./",
            accept_terms=False,
            with_family=False,
        )
        main()

    # Bad username or password
    with (
        patch("getpass.getpass", return_value="invalid_pass"),
        patch("argparse.ArgumentParser.parse_args") as mock_parse_args,
        patch("builtins.open", new_callable=mock_open),
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
        pytest.raises(RuntimeError, match="Bad username or password for invalid_user"),
    ):
        mock_parse_args.return_value = argparse.Namespace(
            username="invalid_user",
            password=None,
            debug=False,
            interactive=True,
            china_mainland=False,
            delete_from_keyring=False,
            loglevel="error",
            no_verify_ssl=True,
            http_proxy=None,
            https_proxy=None,
            session_dir="./",
            accept_terms=False,
            with_family=False,
        )
        main()

    # We should not use getpass for this one, but we reset the password at login fail
    with (
        patch("argparse.ArgumentParser.parse_args") as mock_parse_args,
        patch("builtins.open", new_callable=mock_open),
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
        pytest.raises(RuntimeError, match="Bad username or password for invalid_user"),
    ):
        mock_parse_args.return_value = argparse.Namespace(
            username="invalid_user",
            password="invalid_pass",
            debug=False,
            interactive=False,
            china_mainland=False,
            delete_from_keyring=False,
            loglevel="warning",
            no_verify_ssl=False,
            http_proxy="http://proxy:8080",
            https_proxy="https://proxy:8080",
            session_dir="./",
            accept_terms=True,
            with_family=True,
        )
        main()


def test_username_password_requires_2fa() -> None:
    """Test username and password commands."""
    # Valid connection for the first time
    with (
        patch("argparse.ArgumentParser.parse_args") as mock_parse_args,
        patch("pyicloud.cmdline.input", return_value=VALID_2FA_CODE),
        patch("pyicloud.cmdline.confirm", return_value=False),
        patch("keyring.get_password", return_value=None),
        patch("builtins.open", new_callable=mock_open),
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
    ):
        mock_parse_args.return_value = argparse.Namespace(
            username=REQUIRES_2FA_USER,
            password=VALID_PASSWORD,
            debug=False,
            interactive=True,
            china_mainland=False,
            delete_from_keyring=False,
            device_id=None,
            locate=None,
            output_to_file=None,
            longlist=None,
            list=None,
            sound=None,
            message=None,
            silentmessage=None,
            lostmode=None,
            loglevel="warning",
            no_verify_ssl=True,
            http_proxy=None,
            https_proxy=None,
            session_dir="./",
            accept_terms=False,
            with_family=False,
        )
        main()


def test_device_outputfile(mock_file_open_write_fixture: MagicMock) -> None:
    """Test the outputfile command."""

    with (
        patch("argparse.ArgumentParser.parse_args") as mock_parse_args,
        patch("builtins.open", mock_file_open_write_fixture),
        patch("keyring.get_password", return_value=None),
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
    ):
        mock_parse_args.return_value = argparse.Namespace(
            username=AUTHENTICATED_USER,
            password=VALID_PASSWORD,
            debug=False,
            interactive=False,
            china_mainland=False,
            delete_from_keyring=False,
            device_id=None,
            locate=None,
            output_to_file=True,
            longlist=None,
            list=None,
            sound=None,
            message=None,
            silentmessage=None,
            lostmode=None,
            loglevel="none",
            no_verify_ssl=True,
            http_proxy=None,
            https_proxy=None,
            session_dir="./",
            accept_terms=False,
            with_family=False,
        )
        main()

        devices = FMI_FAMILY_WORKING.get("content")
        if devices:
            for device in devices:
                file_name = device.get("name").strip().lower() + ".fmip_snapshot"
                assert file_name in mock_file_open_write_fixture.written_data
                buffer = BytesIO(mock_file_open_write_fixture.written_data[file_name])

                contents = []
                while True:
                    try:
                        contents.append(pickle.load(buffer))
                    except EOFError:
                        break
                assert contents == [device]


def test_create_pickled_data() -> None:
    """Test the creation of pickled data."""
    idevice = MagicMock()
    idevice.data = {"key": "value"}
    filename = "test.pkl"
    with (
        patch("builtins.open", new_callable=mock_open) as mock_file,
        patch("pickle.dump") as mock_pickle_dump,
        patch("pyicloud.base.PyiCloudSession", new=PyiCloudSessionMock),
    ):
        create_pickled_data(idevice, filename)
        mock_file.assert_called_with(filename, "wb")
        mock_pickle_dump.assert_called_with(
            idevice.data, mock_file(), protocol=pickle.HIGHEST_PROTOCOL
        )


def test_create_parser() -> None:
    """Test the creation of the parser."""
    parser: argparse.ArgumentParser = _create_parser()
    assert isinstance(parser, argparse.ArgumentParser)


def test_enable_lost_mode_option() -> None:
    """Test the enable lost mode option."""
    command_line = MagicMock(
        lostmode=True,
        device_id="123",
        lost_phone="1234567890",
        lost_message="Lost",
        lost_password="pass",
    )
    dev = MagicMock()
    _enable_lost_mode_option(command_line, dev)
    dev.lost_device.assert_called_with(
        number="1234567890", text="Lost", newpasscode="pass"
    )


def test_display_device_message_option() -> None:
    """Test the display device message option."""
    command_line = MagicMock(message="Test Message", device_id="123")
    dev = MagicMock()
    _display_device_message_option(command_line, dev)
    dev.display_message.assert_called_with(
        subject="A Message", message="Test Message", sounds=True
    )


def test_display_device_silent_message_option() -> None:
    """Test the display device silent message option."""
    command_line = MagicMock(silentmessage="Silent Message", device_id="123")
    dev = MagicMock()
    _display_device_silent_message_option(command_line, dev)
    dev.display_message.assert_called_with(
        subject="A Silent Message", message="Silent Message", sounds=False
    )


def test_play_device_sound_option() -> None:
    """Test the play device sound option."""
    command_line = MagicMock(sound=True, device_id="123")
    dev = MagicMock()
    _play_device_sound_option(command_line, dev)
    dev.play_sound.assert_called_once()


def test_handle_2sa() -> None:
    """Test the handle 2sa function."""
    api = MagicMock()
    api.send_verification_code.return_value = True
    api.validate_verification_code.return_value = True
    with (
        patch("pyicloud.cmdline.input", side_effect=["0", "123456"]),
        patch(
            "pyicloud.cmdline._show_devices",
            return_value=[{"deviceName": "Test Device"}],
        ),
    ):
        _handle_2sa(api)

        api.send_verification_code.assert_called_once_with(
            {"deviceName": "Test Device"}
        )
        api.validate_verification_code.assert_called_once_with(
            {"deviceName": "Test Device"},
            "123456",
        )


def test_handle_2fa() -> None:
    """Test the handle 2fa function."""
    api = MagicMock()
    api.validate_2fa_code.return_value = True
    with patch("pyicloud.cmdline.input", return_value="123456"):
        _handle_2fa(api)
        api.validate_2fa_code.assert_called_once_with("123456")


def test_list_devices_option_locate() -> None:
    """Test the list devices option with locate."""
    # Create a mock command_line object with the locate option enabled
    command_line = MagicMock(
        locate=True,  # Enable the locate option
        longlist=False,
        output_to_file=False,
        list=False,
    )

    # Create a mock device object

    dev = MagicMock()
    location = PropertyMock(return_value="Test Location")
    type(dev).location = location

    # Call the function
    _list_devices_option(command_line, dev)

    # Verify that the location() method was called
    location.assert_called_once()


def test_list_devices_option() -> None:
    """Test the list devices option."""
    command_line = MagicMock(
        longlist=True,
        locate=False,
        output_to_file=False,
        list=False,
    )
    content: dict[str, str] = {
        "name": "Test Device",
        "deviceDisplayName": "Test Display",
        "location": "Test Location",
        "batteryLevel": "100%",
        "batteryStatus": "Charging",
        "deviceClass": "Phone",
        "deviceModel": "iPhone",
    }
    dev = AppleDevice(
        content=content,
        params={},
        manager=MagicMock(),
        sound_url="",
        lost_url="",
        message_url="",
        erase_token_url="",
        erase_url="",
    )

    with patch("pyicloud.cmdline.create_pickled_data") as mock_create_pickled:
        _list_devices_option(command_line, dev)

        # Verify no pickled data creation
        mock_create_pickled.assert_not_called()

    # Check for proper console output during detailed listing
    with patch("builtins.print") as mock_print:
        _list_devices_option(command_line, dev)
        mock_print.assert_any_call("-" * 30)
        mock_print.assert_any_call("Test Device")
        for key, value in content.items():
            mock_print.assert_any_call(f"{key:>30} - {pformat(value)}")


def test_list_devices_option_short_list() -> None:
    """Test the list devices option with short list."""
    # Create a mock command_line object with the list option enabled
    command_line = MagicMock(
        longlist=False,
        locate=False,
        output_to_file=False,
        list=True,  # Enable the short list option
    )

    # Create a mock device with sample content
    content: dict[str, str | list[dict[str, bool]]] = {
        "name": "Test Device",
        "deviceDisplayName": "Test Display",
        "location": "Test Location",
        "batteryLevel": "100%",
        "batteryStatus": "Charging",
        "deviceClass": "Phone",
        "deviceModel": "iPhone",
        "features": [
            {"LOC": True},
        ],
    }
    dev = AppleDevice(
        content=content,
        params={},
        manager=MagicMock(),
        sound_url="",
        lost_url="",
        message_url="",
        erase_token_url="",
        erase_url="",
    )

    with patch("builtins.print") as mock_print:
        # Call the function
        _list_devices_option(command_line, dev)

        # Verify the output for short list option
        mock_print.assert_any_call("-" * 30)
        mock_print.assert_any_call("Name           - Test Device")
        mock_print.assert_any_call("Display Name   - Test Display")
        mock_print.assert_any_call("Location       - Test Location")
        mock_print.assert_any_call("Battery Level  - 100%")
        mock_print.assert_any_call("Battery Status - Charging")
        mock_print.assert_any_call("Device Class   - Phone")
        mock_print.assert_any_call("Device Model   - iPhone")



"""The pyiCloud library."""

import logging

from pyicloud.base import AppleDevice, PyiCloudService

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__: list[str] = [
    "PyiCloudService",
    "AppleDevice",
]



"""Contants for the PyiCloud API."""

from enum import IntEnum

CONTENT_TYPE = "Content-Type"
CONTENT_TYPE_JSON = "application/json"
CONTENT_TYPE_TEXT = "plain/text"
CONTENT_TYPE_TEXT_JSON = "text/json"

HEADER_DATA: dict[str, str] = {
    "X-Apple-ID-Account-Country": "account_country",
    "X-Apple-ID-Session-Id": "session_id",
    "X-Apple-Auth-Attributes": "auth_attributes",
    "X-Apple-Session-Token": "session_token",
    "X-Apple-TwoSV-Trust-Token": "trust_token",
    "X-Apple-TwoSV-Trust-Eligible": "trust_eligible",
    "X-Apple-OAuth-Grant-Code": "grant_code",
    "X-Apple-I-Rscd": "apple_rscd",
    "X-Apple-I-Ercd": "apple_ercd",
    "scnt": "scnt",
}

ACCOUNT_NAME = "accountName"


ERROR_ACCESS_DENIED = "ACCESS_DENIED"
ERROR_ZONE_NOT_FOUND = "ZONE_NOT_FOUND"
ERROR_AUTHENTICATION_FAILED = "AUTHENTICATION_FAILED"


class AppleAuthError(IntEnum):
    """Apple auth error codes."""

    SUCCESS = 200
    LOGIN_TOKEN_EXPIRED = 421
    TWO_FACTOR_REQUIRED = 409
    FIND_MY_REAUTH_REQUIRED = 450
    GENERAL_AUTH_ERROR = 500



"""Account test constants."""

# pylint: disable=line-too-long
from .const_login import FIRST_NAME

# Fakers
PAYMENT_METHOD_ID_1 = "PAYMENT_METHOD_ID_1"
PAYMENT_METHOD_ID_2 = "PAYMENT_METHOD_ID_2"
PAYMENT_METHOD_ID_3 = "PAYMENT_METHOD_ID_3"
PAYMENT_METHOD_ID_4 = "PAYMENT_METHOD_ID_4"

# Data
ACCOUNT_DEVICES_WORKING = {
    "devices": [
        {
            "serialNumber": "●●●●●●●NG123",
            "osVersion": "OSX;10.15.3",
            "modelLargePhotoURL2x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/MacBookPro/MacBookPro15,1-spacegray/online-infobox__2x.png",
            "modelLargePhotoURL1x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/MacBookPro/MacBookPro15,1-spacegray/online-infobox.png",
            "paymentMethods": [PAYMENT_METHOD_ID_3],
            "name": "MacBook Pro de " + FIRST_NAME,
            "imei": "",
            "model": "MacBookPro15,1",
            "udid": "MacBookPro15,1" + FIRST_NAME,
            "modelSmallPhotoURL2x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/MacBookPro/MacBookPro15,1-spacegray/online-sourcelist__2x.png",
            "modelSmallPhotoURL1x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/MacBookPro/MacBookPro15,1-spacegray/online-sourcelist.png",
            "modelDisplayName": 'MacBook Pro 15"',
        },
        {
            "serialNumber": "●●●●●●●UX123",
            "osVersion": "iOS;13.3",
            "modelLargePhotoURL2x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/iPhone/iPhone12,1-1-6-0/online-infobox__2x.png",
            "modelLargePhotoURL1x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/iPhone/iPhone12,1-1-6-0/online-infobox.png",
            "paymentMethods": [
                PAYMENT_METHOD_ID_4,
                PAYMENT_METHOD_ID_2,
                PAYMENT_METHOD_ID_1,
            ],
            "name": "iPhone de " + FIRST_NAME,
            "imei": "●●●●●●●●●●12345",
            "model": "iPhone12,1",
            "udid": "iPhone12,1" + FIRST_NAME,
            "modelSmallPhotoURL2x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/iPhone/iPhone12,1-1-6-0/online-sourcelist__2x.png",
            "modelSmallPhotoURL1x": "https://statici.icloud.com/fmipmobile/deviceImages-4.0/iPhone/iPhone12,1-1-6-0/online-sourcelist.png",
            "modelDisplayName": "iPhone 11",
        },
    ],
    "paymentMethods": [
        {
            "lastFourDigits": "333",
            "balanceStatus": "NOTAPPLICABLE",
            "suspensionReason": "ACTIVE",
            "id": PAYMENT_METHOD_ID_3,
            "type": "Boursorama Banque",
        },
        {
            "lastFourDigits": "444",
            "balanceStatus": "NOTAPPLICABLE",
            "suspensionReason": "ACTIVE",
            "id": PAYMENT_METHOD_ID_4,
            "type": "Carte Crédit Agricole",
        },
        {
            "lastFourDigits": "2222",
            "balanceStatus": "NOTAPPLICABLE",
            "suspensionReason": "ACTIVE",
            "id": PAYMENT_METHOD_ID_2,
            "type": "Lydia",
        },
        {
            "lastFourDigits": "111",
            "balanceStatus": "NOTAPPLICABLE",
            "suspensionReason": "ACTIVE",
            "id": PAYMENT_METHOD_ID_1,
            "type": "Boursorama Banque",
        },
    ],
}


ACCOUNT_STORAGE_WORKING = {
    "storageUsageByMedia": [
        {
            "mediaKey": "photos",
            "displayLabel": "Photos et vidéos",
            "displayColor": "ffcc00",
            "usageInBytes": 0,
        },
        {
            "mediaKey": "backup",
            "displayLabel": "Sauvegarde",
            "displayColor": "5856d6",
            "usageInBytes": 799008186,
        },
        {
            "mediaKey": "docs",
            "displayLabel": "Documents",
            "displayColor": "ff9500",
            "usageInBytes": 449092146,
        },
        {
            "mediaKey": "mail",
            "displayLabel": "Mail",
            "displayColor": "007aff",
            "usageInBytes": 1101522944,
        },
    ],
    "storageUsageInfo": {
        "compStorageInBytes": 0,
        "usedStorageInBytes": 2348632876,
        "totalStorageInBytes": 5368709120,
        "commerceStorageInBytes": 0,
    },
    "quotaStatus": {
        "overQuota": False,
        "haveMaxQuotaTier": False,
        "almost-full": False,
        "paidQuota": False,
    },
}



"""Utils."""

import base64
import getpass
import sys
from typing import Optional

import keyring

KEYRING_SYSTEM = "pyicloud://icloud-password"


def get_password(username: str, interactive=sys.stdout.isatty()) -> Optional[str]:
    """Get the password from a username.
    Returns the password if found in keyring or if interactive is True.
    Returns None if no password is found and interactive is False."""
    result: Optional[str] = get_password_from_keyring(username)
    if result:
        return result

    if interactive:
        return getpass.getpass(f"Enter iCloud password for {username}: ")


def password_exists_in_keyring(username: str) -> bool:
    """Return true if the password of a username exists in the keyring."""
    return get_password_from_keyring(username) is not None


def get_password_from_keyring(username: str) -> Optional[str]:
    """Get the password from a username."""
    return keyring.get_password(KEYRING_SYSTEM, username)


def store_password_in_keyring(username: str, password: str) -> None:
    """Store the password of a username."""
    return keyring.set_password(
        KEYRING_SYSTEM,
        username,
        password,
    )


def delete_password_in_keyring(username: str) -> None:
    """Delete the password of a username."""
    return keyring.delete_password(
        KEYRING_SYSTEM,
        username,
    )


def underscore_to_camelcase(word: str, initial_capital: bool = False) -> str:
    """Transform a word to camelCase."""
    words: list[str] = [x.capitalize() or "_" for x in word.split("_")]
    if not initial_capital:
        words[0] = words[0].lower()

    return "".join(words)


def camelcase_to_underscore(camel_str: str) -> str:
    """
    Convert camelCase string to snake_case.

    Examples:
        startDate -> start_date
        localStartDate -> local_start_date
        hasAttachments -> has_attachments
    """
    result: list[str] = []
    for i, char in enumerate(camel_str):
        if char.isupper() and i > 0:
            result.append("_")
        result.append(char.lower())
    return "".join(result)


def b64url_decode(s: str) -> bytes:
    """Decode a base64url encoded string."""
    return base64.urlsafe_b64decode(s + "=" * (-len(s) % 4))


def b64_encode(b: bytes) -> str:
    """Encode bytes to a base64 encoded string."""
    return base64.b64encode(b).decode()



"""Tests for the PyiCloudCookieJar class and its handling of FMIP auth cookies."""

from io import StringIO
from unittest.mock import MagicMock, mock_open, patch

from pyicloud.cookie_jar import _FMIP_AUTH_COOKIE_NAME, PyiCloudCookieJar


def create_cookie_jar_with_cookie(
    filename, name, domain="example.com", path="/", value="test"
) -> PyiCloudCookieJar:
    """Create a PyiCloudCookieJar with a single cookie."""
    with (
        patch("builtins.open", new_callable=mock_open),
        patch("os.open", new_callable=mock_open),
    ):
        jar = PyiCloudCookieJar(filename=filename)
        jar.set(name, value, domain=domain, path=path)
        return jar


def test_load_no_filename() -> None:
    """Test that load is a no-op if no filename is set."""
    jar = PyiCloudCookieJar()
    # Should not raise or do anything if no filename is set
    # with patch("builtins.open", mock_open()):
    jar.load()  # No-op


def test_load_with_filename_removes_fmip_cookie() -> None:
    """Test that loading a jar with an FMIP cookie removes that cookie."""
    filename = "test_cookies.txt"
    buffer = StringIO()
    buffer.close = MagicMock()
    with (
        patch("builtins.open", new_callable=mock_open) as m,
        patch("os.open"),
        patch("os.fdopen") as os_fdopen,
    ):
        m.return_value = buffer
        os_fdopen.return_value = buffer

        jar: PyiCloudCookieJar = create_cookie_jar_with_cookie(
            filename, _FMIP_AUTH_COOKIE_NAME
        )
        # Add a non-FMIP cookie too
        jar.set("other_cookie", "value", domain="example.com", path="/")
        jar.save()
        # Reload and check FMIP cookie is removed
        jar2 = PyiCloudCookieJar(filename=filename)
        buffer.seek(0)
        jar2.load()
        names: list[str] = [cookie.name for cookie in jar2]
        assert _FMIP_AUTH_COOKIE_NAME not in names
        assert "other_cookie" in names


def test_load_with_custom_filename_argument_removes_fmip_cookie() -> None:
    """Test that loading a jar with an FMIP cookie removes that cookie."""
    filename = "test_cookies.txt"
    buffer = StringIO()
    buffer.close = MagicMock()
    with (
        patch("builtins.open", new_callable=mock_open) as m,
        patch("os.open"),
        patch("os.fdopen") as os_fdopen,
    ):
        m.return_value = buffer
        os_fdopen.return_value = buffer
        jar: PyiCloudCookieJar = create_cookie_jar_with_cookie(
            filename, _FMIP_AUTH_COOKIE_NAME
        )
        jar.save()

        jar2: PyiCloudCookieJar = PyiCloudCookieJar()
        names: list[str] = [cookie.name for cookie in jar2]
        assert _FMIP_AUTH_COOKIE_NAME not in names


def test_load_handles_keyerror_on_clear() -> None:
    """Test that load handles KeyError from clear gracefully."""
    filename = "test_cookies.txt"
    buffer = StringIO()
    buffer.close = MagicMock()
    with (
        patch("builtins.open", new_callable=mock_open) as m,
        patch("os.open"),
        patch("os.fdopen") as os_fdopen,
    ):
        m.return_value = buffer
        os_fdopen.return_value = buffer
        jar: PyiCloudCookieJar = create_cookie_jar_with_cookie(
            filename, _FMIP_AUTH_COOKIE_NAME
        )
        jar.save()
        buffer.seek(0)
        jar2: PyiCloudCookieJar = PyiCloudCookieJar(filename=filename)

        # Monkeypatch clear to raise KeyError
        def raise_keyerror(*args, **kwargs) -> None:
            raise KeyError

        with patch.object(jar2, "clear", side_effect=raise_keyerror):
            # Should not raise
            jar2.load()



"""Account service tests."""

# pylint: disable=protected-access

from unittest.mock import MagicMock

from pyicloud import PyiCloudService
from pyicloud.services.account import AccountStorageUsage


def test_repr(pyicloud_service_working: PyiCloudService) -> None:
    """Tests representation."""
    assert (
        repr(pyicloud_service_working.account)
        == "<AccountService: {devices: 2, family: 3, storage: 3020076244 bytes free}>"
    )


def test_devices(pyicloud_service_working: PyiCloudService) -> None:
    """Tests devices."""
    assert pyicloud_service_working.account.devices
    assert len(pyicloud_service_working.account.devices) == 2

    for device in pyicloud_service_working.account.devices:
        assert device.name
        assert device.model
        assert device.udid
        assert device["serialNumber"]
        assert device["osVersion"]
        assert device["modelLargePhotoURL2x"]
        assert device["modelLargePhotoURL1x"]
        assert device["paymentMethods"]
        assert device["name"]
        assert device["model"]
        assert device["udid"]
        assert device["modelSmallPhotoURL2x"]
        assert device["modelSmallPhotoURL1x"]
        assert device["modelDisplayName"]
        assert (
            repr(device)
            == "<AccountDevice: {model: "
            + device.model_display_name
            + ", name: "
            + device.name
            + "}>"
        )


def test_family(pyicloud_service_working: PyiCloudService) -> None:
    """Tests family members."""
    assert pyicloud_service_working.account.family
    assert len(pyicloud_service_working.account.family) == 3

    for member in pyicloud_service_working.account.family:
        assert member.last_name
        assert member.dsid
        assert member.original_invitation_email
        assert member.full_name
        assert member.age_classification
        assert member.apple_id_for_purchases
        assert member.apple_id
        assert member.first_name
        assert not member.has_screen_time_enabled
        assert not member.has_ask_to_buy_enabled
        assert not member.share_my_location_enabled_family_members
        assert member.dsid_for_purchases
        assert (
            repr(member)
            == "<FamilyMember: {name: "
            + member.full_name
            + ", age_classification: "
            + member.age_classification
            + "}>"
        )


def test_storage(pyicloud_service_working: PyiCloudService) -> None:
    """Tests storage."""
    assert pyicloud_service_working.account.storage
    assert repr(pyicloud_service_working.account.storage) == (
        "<AccountStorage: {usage: 43.75% used of 5368709120 bytes, usages_by_media: "
        "{'photos': <AccountStorageUsageForMedia: {key: photos, usage: 0 bytes}>, "
        "'backup': <AccountStorageUsageForMedia: {key: backup, usage: 799008186 bytes}>, "
        "'docs': <AccountStorageUsageForMedia: {key: docs, usage: 449092146 bytes}>, "
        "'mail': <AccountStorageUsageForMedia: {key: mail, usage: 1101522944 bytes}>}}>"
    )


def test_storage_usage(pyicloud_service_working: PyiCloudService) -> None:
    """Tests storage usage."""
    assert pyicloud_service_working.account.storage.usage
    usage: AccountStorageUsage = pyicloud_service_working.account.storage.usage
    assert usage.comp_storage_in_bytes or usage.comp_storage_in_bytes == 0
    assert usage.used_storage_in_bytes
    assert usage.used_storage_in_percent
    assert usage.available_storage_in_bytes
    assert usage.available_storage_in_percent
    assert usage.total_storage_in_bytes
    assert usage.commerce_storage_in_bytes or usage.commerce_storage_in_bytes == 0
    assert not usage.quota_over
    assert not usage.quota_tier_max
    assert not usage.quota_almost_full
    assert not usage.quota_paid
    assert (
        repr(usage)
        == "<AccountStorageUsage: "
        + str(usage.used_storage_in_percent)
        + "% used of "
        + str(usage.total_storage_in_bytes)
        + " bytes>"
    )


def test_storage_usages_by_media(pyicloud_service_working: PyiCloudService) -> None:
    """Tests storage usages by media."""
    assert pyicloud_service_working.account.storage.usages_by_media

    for (
        usage_media
    ) in pyicloud_service_working.account.storage.usages_by_media.values():
        assert usage_media.key
        assert usage_media.label
        assert usage_media.color
        assert usage_media.usage_in_bytes or usage_media.usage_in_bytes == 0
        assert (
            repr(usage_media)
            == "<AccountStorageUsageForMedia: {key: "
            + usage_media.key
            + ", usage: "
            + str(usage_media.usage_in_bytes)
            + " bytes}>"
        )


def test_summary_plan(
    pyicloud_service_working: PyiCloudService, mock_session: MagicMock
) -> None:
    """Tests the summary_plan property."""
    # Mock the response for the summary plan endpoint
    mock_response = {
        "planName": "iCloud+",
        "storageCapacity": "200GB",
        "price": "$2.99/month",
    }
    mock_session.get.return_value.json.return_value = mock_response
    pyicloud_service_working._session = mock_session

    # Access the summary_plan property
    summary_plan = pyicloud_service_working.account.summary_plan

    # Assertions
    assert summary_plan == mock_response
    mock_session.get.assert_called_once_with(
        pyicloud_service_working.account._gateway_summary_plan_url,
        params=pyicloud_service_working.account.params,
    )



"""Library exceptions."""

from typing import Optional, Union

from requests import Response


class PyiCloudException(Exception):
    """Generic iCloud exception."""


class PyiCloudPasswordException(PyiCloudException):
    """Password exception."""


class PyiCloudServiceUnavailable(PyiCloudException):
    """Service unavailable exception."""


class TokenException(PyiCloudException):
    """Token exception."""


# API
class PyiCloudAPIResponseException(PyiCloudException):
    """iCloud response exception."""

    def __init__(
        self,
        reason: str,
        code: Optional[Union[int, str]] = None,
        response: Optional[Response] = None,
    ) -> None:
        self.reason: str = reason
        self.code: Optional[Union[int, str]] = code
        self.response: Optional[Response] = response
        message: str = reason or ""
        if code:
            message += f" ({code})"

        if response is not None and response.text:
            message += f": {response.text}"

        super().__init__(message)


class PyiCloudServiceNotActivatedException(PyiCloudAPIResponseException):
    """iCloud service not activated exception."""


# Login
class PyiCloudFailedLoginException(PyiCloudException):
    """iCloud failed login exception."""

    def __init__(
        self,
        msg: str,
        *args,
        response: Optional[Response] = None,
    ) -> None:
        self.response: Optional[Response] = response
        message: str = msg or "Failed login to iCloud"
        if response is not None and response.text:
            message = f"{message} ({response.status_code}): {response.text}"
        super().__init__(message, *args)


class PyiCloudAcceptTermsException(PyiCloudException):
    """iCloud accept terms exception."""


class PyiCloud2FARequiredException(PyiCloudException):
    """iCloud 2FA required exception."""

    def __init__(self, apple_id: str, response: Response) -> None:
        message: str = f"2FA authentication required for account: {apple_id} (HSA2)"
        super().__init__(message)
        self.response: Response = response


class PyiCloud2SARequiredException(PyiCloudException):
    """iCloud 2SA required exception."""

    def __init__(self, apple_id: str) -> None:
        message: str = f"Two-step authentication required for account: {apple_id}"
        super().__init__(message)


class PyiCloudAuthRequiredException(PyiCloudException):
    """iCloud re-authentication required exception."""

    def __init__(self, apple_id: str, response: Response) -> None:
        message: str = f"Re-authentication required for account: {apple_id}"
        super().__init__(message)
        self.response: Response = response


class PyiCloudNoTrustedNumberAvailable(PyiCloudException):
    """iCloud no trusted number exception."""


class PyiCloudNoStoredPasswordAvailableException(PyiCloudException):
    """iCloud no stored password exception."""


# Webservice specific
class PyiCloudNoDevicesException(PyiCloudException):
    """iCloud no device exception."""



"""Unit tests for the RemindersService class."""
# pylint: disable=protected-access

import datetime
from unittest.mock import MagicMock, patch

from requests import Response

from pyicloud.services.reminders import RemindersService
from pyicloud.session import PyiCloudSession


def test_reminders_service_init(mock_session: MagicMock) -> None:
    """Test RemindersService initialization."""
    mock_session.get.return_value = MagicMock(
        spec=Response, json=lambda: {"Collections": [], "Reminders": []}
    )
    params: dict[str, str] = {"dsid": "12345"}

    with patch("pyicloud.services.reminders.get_localzone_name", return_value="UTC"):
        service = RemindersService("https://example.com", mock_session, params)

        assert service.service_root == "https://example.com"
        assert service.params == params
        assert not service.lists
        assert not service.collections


def test_reminders_service_refresh() -> None:
    """Test the refresh method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {
        "Collections": [
            {"title": "Work", "guid": "guid1", "ctag": "ctag1"},
            {"title": "Personal", "guid": "guid2", "ctag": "ctag2"},
        ],
        "Reminders": [
            {"title": "Task 1", "pGuid": "guid1", "dueDate": [2023, 10, 1, 12, 0, 0]},
            {"title": "Task 2", "pGuid": "guid2", "dueDate": None},
        ],
    }
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.reminders.get_localzone_name", return_value="UTC"):
        service = RemindersService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.refresh()

        assert "Work" in service.lists
        assert "Personal" in service.lists
        assert len(service.lists["Work"]) == 1
        assert len(service.lists["Personal"]) == 1

        work_task = service.lists["Work"][0]
        assert work_task["title"] == "Task 1"
        assert work_task["due"] == datetime.datetime(2023, 10, 1, 12, 0, 0)

        personal_task = service.lists["Personal"][0]
        assert personal_task["title"] == "Task 2"
        assert personal_task["due"] is None


def test_reminders_service_post() -> None:
    """Test the post method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.ok = True
    mock_session.post.return_value = mock_response

    with patch("pyicloud.services.reminders.get_localzone_name", return_value="UTC"):
        service = RemindersService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.collections = {"Work": {"guid": "guid1"}}

        # Test posting a reminder with a due date
        due_date = datetime.datetime(2023, 10, 1, 12, 0, 0)
        result: bool = service.post("New Task", "Description", "Work", due_date)

        assert result is True
        mock_session.post.assert_called_once()
        _, kwargs = mock_session.post.call_args
        assert kwargs["json"]
        data = kwargs["json"]
        assert data["Reminders"]["title"] == "New Task"
        assert data["Reminders"]["description"] == "Description"
        assert data["Reminders"]["pGuid"] == "guid1"
        assert data["Reminders"]["dueDate"] == [20231001, 2023, 10, 1, 12, 0]

        # Test posting a reminder without a due date
        mock_session.post.reset_mock()
        result = service.post("Task Without Due Date", collection="Work")

        assert result is True
        mock_session.post.assert_called_once()
        _, kwargs = mock_session.post.call_args
        data = kwargs["json"]
        assert data["Reminders"]["title"] == "Task Without Due Date"
        assert data["Reminders"]["dueDate"] is None


def test_reminders_service_post_invalid_collection() -> None:
    """Test the post method with an invalid collection."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.ok = True
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.reminders.get_localzone_name", return_value="UTC"):
        service = RemindersService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )

        # Post to a non-existent collection
        result = service.post("Task", collection="NonExistent")
        assert result is True
        mock_session.post.assert_called_once()
        _, kwargs = mock_session.post.call_args
        data = kwargs["json"]
        assert data["Reminders"]["pGuid"] == "tasks"  # Default collection


def test_reminders_service_refresh_empty_response() -> None:
    """Test the refresh method with an empty response."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Collections": [], "Reminders": []}
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.reminders.get_localzone_name", return_value="UTC"):
        service = RemindersService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.refresh()

        assert not service.lists
        assert not service.collections



"""Cookie jar with persistence support."""

from http.cookiejar import LWPCookieJar
from typing import Optional

from requests.cookies import RequestsCookieJar

_FMIP_AUTH_COOKIE_NAME: str = "X-APPLE-WEBAUTH-FMIP"


class PyiCloudCookieJar(RequestsCookieJar, LWPCookieJar):
    """Mix the Requests CookieJar with the LWPCookieJar to allow persistance"""

    def __init__(self, filename: Optional[str] = None) -> None:
        """Initialise both bases; do not pass filename positionally to RequestsCookieJar."""
        RequestsCookieJar.__init__(self)
        LWPCookieJar.__init__(self, filename=filename)

    def _resolve_filename(self, filename: Optional[str]) -> Optional[str]:
        resolved: Optional[str] = filename or getattr(self, "filename", None)
        if not resolved:
            return  # No-op if no filename is bound
        return resolved

    def load(
        self,
        filename: Optional[str] = None,
        ignore_discard: bool = True,
        ignore_expires: bool = False,
    ) -> None:
        """Load cookies from file."""
        resolved: Optional[str] = self._resolve_filename(filename)
        if not resolved:
            return  # No-op if no filename is bound
        super().load(
            filename=resolved,
            ignore_discard=ignore_discard,
            ignore_expires=ignore_expires,
        )
        # Clear any FMIP cookie regardless of domain/path to avoid stale auth.
        cookies_to_clear: list[tuple[str, str, str]] = [
            (cookie.domain, cookie.path, cookie.name)
            for cookie in self
            if cookie.name == _FMIP_AUTH_COOKIE_NAME
        ]
        for domain, path, name in cookies_to_clear:
            try:
                self.clear(domain=domain, path=path, name=name)
            except KeyError:
                pass

    def save(
        self,
        filename: Optional[str] = None,
        ignore_discard: bool = True,
        ignore_expires: bool = False,
    ) -> None:
        """Save cookies to file."""
        resolved: Optional[str] = self._resolve_filename(filename)
        if not resolved:
            return  # No-op if no filename is bound
        super().save(
            filename=resolved,
            ignore_discard=ignore_discard,
            ignore_expires=ignore_expires,
        )



"""Drive service."""

import io
import logging
import mimetypes
import os
import time
import uuid
from datetime import datetime, timedelta
from re import Match, search
from typing import IO, Any, Optional

from requests import Response

from pyicloud.const import CONTENT_TYPE, CONTENT_TYPE_TEXT
from pyicloud.exceptions import PyiCloudAPIResponseException, TokenException
from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession

LOGGER: logging.Logger = logging.getLogger(__name__)

COOKIE_APPLE_WEBAUTH_VALIDATE: str = "X-APPLE-WEBAUTH-VALIDATE"
CLOUD_DOCS_ZONE: str = "com.apple.CloudDocs"
NODE_ROOT: str = "root"
NODE_TRASH: str = "TRASH_ROOT"
CLOUD_DOCS_ZONE_ID_ROOT: str = f"FOLDER::{CLOUD_DOCS_ZONE}::{NODE_ROOT}"
CLOUD_DOCS_ZONE_ID_TRASH: str = f"FOLDER::{CLOUD_DOCS_ZONE}::{NODE_TRASH}"


class DriveService(BaseService):
    """The 'Drive' iCloud service."""

    def __init__(
        self,
        service_root: str,
        document_root: str,
        session: PyiCloudSession,
        params: dict[str, Any],
    ) -> None:
        super().__init__(service_root, session, params)
        self._document_root: str = document_root
        self._root: Optional[DriveNode] = None
        self._trash: Optional[DriveNode] = None

    def _get_token_from_cookie(self) -> dict[str, Any]:
        for cookie in self.session.cookies:
            if cookie.name == COOKIE_APPLE_WEBAUTH_VALIDATE and cookie.value:
                match: Optional[Match[str]] = search(r"\bt=([^:]+)", cookie.value)
                if match is None:
                    raise TokenException(f"Can't extract token from {cookie.value}")
                return {"token": match.group(1)}
        raise TokenException("Token cookie not found")

    def get_node_data(
        self, drivewsid: str, share_id: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """Returns the node data."""
        payload = [
            {
                "drivewsid": drivewsid,
                "partialData": False,
            }
        ]
        if share_id:
            payload[0]["shareID"] = share_id
        request: Response = self.session.post(
            self.service_root + "/retrieveItemDetailsInFolders",
            params=self.params,
            json=payload,
        )
        self._raise_if_error(request)
        return request.json()[0]

    def get_file(self, file_id: str, zone: str = CLOUD_DOCS_ZONE, **kwargs) -> Response:
        """Returns iCloud Drive file."""
        file_params: dict[str, Any] = dict(self.params)
        file_params.update({"document_id": file_id})
        response: Response = self.session.get(
            self._document_root + f"/ws/{zone}/download/by_id",
            params=file_params,
        )
        self._raise_if_error(response)
        response_json = response.json()
        package_token = response_json.get("package_token")
        data_token = response_json.get("data_token")
        if data_token and data_token.get("url"):
            return self.session.get(data_token["url"], params=self.params, **kwargs)
        if package_token and package_token.get("url"):
            return self.session.get(package_token["url"], params=self.params, **kwargs)
        raise KeyError("'data_token' nor 'package_token'")

    def get_app_data(self):
        """Returns the app library (previously ubiquity)."""
        request: Response = self.session.get(
            self.service_root + "/retrieveAppLibraries",
            params=self.params,
        )
        self._raise_if_error(request)
        return request.json()["items"]

    def _get_upload_contentws_url(
        self,
        file_object: IO,
        zone: str = CLOUD_DOCS_ZONE,
    ) -> tuple[str, str]:
        """Get the contentWS endpoint URL to add a new file."""

        content_type: Optional[str] = mimetypes.guess_type(file_object.name)[0]
        if content_type is None:
            content_type = ""

        # Get filesize from file object
        orig_pos: int = file_object.tell()
        file_object.seek(0, os.SEEK_END)
        file_size: int = file_object.tell()
        file_object.seek(orig_pos, os.SEEK_SET)

        file_params: dict[str, Any] = self.params
        file_params.update(self._get_token_from_cookie())

        request: Response = self.session.post(
            self._document_root + f"/ws/{zone}/upload/web",
            params=file_params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "filename": file_object.name,
                "type": "FILE",
                "content_type": content_type,
                "size": file_size,
            },
        )
        self._raise_if_error(request)
        return request.json()[0]["document_id"], request.json()[0]["url"]

    def _update_contentws(
        self,
        folder_id: str,
        file_info: dict[str, Any],
        document_id: str,
        file_object: IO,
        zone: str = CLOUD_DOCS_ZONE,
        **kwargs,
    ):
        data: dict[str, Any] = {
            "data": {
                "signature": file_info["fileChecksum"],
                "wrapping_key": file_info["wrappingKey"],
                "reference_signature": file_info["referenceChecksum"],
                "size": file_info["size"],
            },
            "command": "add_file",
            "create_short_guid": True,
            "document_id": document_id,
            "path": {
                "starting_document_id": folder_id,
                "path": os.path.basename(file_object.name),
            },
            "allow_conflict": True,
            "file_flags": {
                "is_writable": True,
                "is_executable": False,
                "is_hidden": False,
            },
            "mtime": int(kwargs.get("mtime", time.time()) * 1000),
            "btime": int(kwargs.get("ctime", time.time()) * 1000),
        }

        # Add the receipt if we have one. Will be absent for 0-sized files
        if file_info.get("receipt"):
            data["data"].update({"receipt": file_info["receipt"]})

        request: Response = self.session.post(
            self._document_root + f"/ws/{zone}/update/documents",
            params=self.params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json=data,
        )
        self._raise_if_error(request)
        return request.json()

    def send_file(
        self,
        folder_id: str,
        file_object: IO,
        zone: str = CLOUD_DOCS_ZONE,
        **kwargs,
    ) -> None:
        """Send new file to iCloud Drive."""
        document_id, content_url = self._get_upload_contentws_url(
            file_object=file_object, zone=zone
        )

        request: Response = self.session.post(
            content_url, files={file_object.name: file_object}
        )
        self._raise_if_error(request)
        content_response = request.json()["singleFile"]
        self._update_contentws(
            folder_id,
            content_response,
            document_id,
            file_object,
            zone,
            **kwargs,
        )

    def create_folders(self, parent: str, name: str):
        """Creates a new iCloud Drive folder"""
        # when creating a folder on icloud.com, the clientID is set to the following:
        temp_client_id: str = f"FOLDER::UNKNOWN_ZONE::TempId-{uuid.uuid4()}"
        request: Response = self.session.post(
            self.service_root + "/createFolders",
            params=self.params,
            headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
            json={
                "destinationDrivewsId": parent,
                "folders": [
                    {
                        "clientId": temp_client_id,
                        "name": name,
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    def delete_items(self, node_id: str, etag: str):
        """Deletes an iCloud Drive node"""
        request: Response = self.session.post(
            self.service_root + "/deleteItems",
            params=self.params,
            json={
                "items": [
                    {
                        "drivewsid": node_id,
                        "etag": etag,
                        "clientId": self.params["clientId"],
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    def rename_items(self, node_id: str, etag: str, name: str):
        """Renames an iCloud Drive node"""
        request: Response = self.session.post(
            self.service_root + "/renameItems",
            params=self.params,
            json={
                "items": [
                    {
                        "drivewsid": node_id,
                        "etag": etag,
                        "name": name,
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    def move_items_to_trash(self, node_id: str, etag: str):
        """Moves an iCloud Drive node to the trash bin"""
        # when moving a node to the trash on icloud.com, the clientID is set to the node_id:
        temp_client_id: str = node_id
        request: Response = self.session.post(
            self.service_root + "/moveItemsToTrash",
            params=self.params,
            json={
                "items": [
                    {
                        "drivewsid": node_id,
                        "etag": etag,
                        "clientId": temp_client_id,
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    def recover_items_from_trash(self, node_id: str, etag: str):
        """Restores an iCloud Drive node from the trash bin"""
        request: Response = self.session.post(
            self.service_root + "/putBackItemsFromTrash",
            params=self.params,
            json={
                "items": [
                    {
                        "drivewsid": node_id,
                        "etag": etag,
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    def delete_forever_from_trash(self, node_id: str, etag: str):
        """Permanently deletes an iCloud Drive node from the trash bin"""
        request: Response = self.session.post(
            self.service_root + "/deleteItems",
            params=self.params,
            json={
                "items": [
                    {
                        "drivewsid": node_id,
                        "etag": etag,
                    }
                ],
            },
        )
        self._raise_if_error(request)
        return request.json()

    @property
    def root(self) -> "DriveNode":
        """Returns the root node."""
        if not self._root:
            self.refresh_root()
        if self._root:
            return self._root
        raise ValueError("Root not found")

    @property
    def trash(self) -> "DriveNode":
        """Returns the trash node."""
        if not self._trash:
            self.refresh_trash()
        if self._trash:
            return self._trash
        raise ValueError("Trash not found")

    def refresh_root(self) -> None:
        """Refreshes and returns a fresh root node."""
        self._root = DriveNode(self, self.get_node_data(CLOUD_DOCS_ZONE_ID_ROOT))

    def refresh_trash(self) -> None:
        """Refreshes and returns a fresh trash node."""
        self._trash = DriveNode(self, self.get_node_data(CLOUD_DOCS_ZONE_ID_TRASH))

    def __getattr__(self, attr):
        return getattr(self.root, attr)

    def __getitem__(self, key: str) -> "DriveNode":
        return self.root[key]

    @staticmethod
    def _raise_if_error(response: Response) -> None:
        if not response.ok:
            api_error = PyiCloudAPIResponseException(
                response.reason, response.status_code
            )
            LOGGER.error(api_error)
            raise api_error


class DriveNode:
    """Drive node."""

    TYPE_UNKNOWN = "unknown"
    TYPE_TRASH = "trash"
    NAME_ROOT = "root"
    NAME_UNKNOWN = "<UNKNOWN>"

    def __init__(self, conn: DriveService, data: dict[str, Any]) -> None:
        self.data: dict[str, Any] = data
        self.connection: DriveService = conn
        self._children: Optional[list[DriveNode]] = None

    @property
    def name(self) -> str:
        """Gets the node name."""
        # check if name is undefined, return drivewsid instead if so.
        node_name: Optional[str] = self.data.get("name")
        if not node_name:
            # use drivewsid as name if no name present.
            node_name = self.data.get("drivewsid")
            # Clean up well-known drivewsid names
            if node_name == CLOUD_DOCS_ZONE_ID_ROOT:
                node_name = self.NAME_ROOT
            # if no name still, return unknown string.
            if not node_name:
                node_name = self.NAME_UNKNOWN

        if "extension" in self.data:
            return f"{node_name}.{self.data['extension']}"
        return node_name

    @property
    def type(self) -> str:
        """Gets the node type."""
        node_type: Optional[str] = self.data.get("type")
        # handle trash which has no node type
        if not node_type and self.data.get("drivewsid") == NODE_TRASH:
            node_type = self.TYPE_TRASH

        if not node_type:
            node_type = self.TYPE_UNKNOWN

        return node_type.lower()

    def get_children(self, force: bool = False) -> list["DriveNode"]:
        """Gets the node children."""
        if not self._children or force:
            if "items" not in self.data or force:
                self.data.update(
                    self.connection.get_node_data(
                        self.data["drivewsid"], self.data.get("shareID")
                    )
                )
            if "items" not in self.data:
                raise KeyError(f"No items in folder, status: {self.data['status']}")
            self._children = [
                DriveNode(self.connection, item_data)
                for item_data in self.data["items"]
            ]
        return self._children

    def remove(self, child: "DriveNode") -> None:
        """Removes a child from the node."""
        if self._children:
            for item_data in self.data["items"]:
                if item_data["docwsid"] == child.data["docwsid"]:
                    self.data["items"].remove(item_data)
                    break
            self._children.remove(child)
        else:
            raise ValueError("No children to remove")

    @property
    def size(self) -> Optional[int]:
        """Gets the node size."""
        size: Optional[str] = self.data.get("size")  # Folder does not have size
        if not size:
            return None
        return int(size)

    @property
    def date_changed(self) -> Optional[datetime]:
        """Gets the node changed date (in UTC)."""
        return _date_to_utc(self.data.get("dateChanged"))  # Folder does not have date

    @property
    def date_modified(self) -> Optional[datetime]:
        """Gets the node modified date (in UTC)."""
        return _date_to_utc(self.data.get("dateModified"))  # Folder does not have date

    @property
    def date_last_open(self) -> Optional[datetime]:
        """Gets the node last open date (in UTC)."""
        return _date_to_utc(self.data.get("lastOpenTime"))  # Folder does not have date

    def open(self, **kwargs):
        """Gets the node file."""
        # iCloud returns 400 Bad Request for 0-byte files
        if self.data["size"] == 0:
            response = Response()
            response.raw = io.BytesIO()
            return response
        return self.connection.get_file(
            self.data["docwsid"], zone=self.data["zone"], **kwargs
        )

    def upload(self, file_object, **kwargs):
        """Upload a new file."""
        return self.connection.send_file(
            self.data["docwsid"], file_object, zone=self.data["zone"], **kwargs
        )

    def dir(self) -> list[str]:
        """Gets the node list of directories."""
        if self.type == "file":
            raise NotADirectoryError(self.name)
        return [child.name for child in self.get_children()]

    def mkdir(self, folder: str):
        """Create a new directory directory."""
        return self.connection.create_folders(self.data["drivewsid"], folder)

    def rename(self, name: str):
        """Rename an iCloud Drive item."""
        return self.connection.rename_items(
            self.data["drivewsid"], self.data["etag"], name
        )

    def move_to_trash(self):
        """Move an iCloud Drive item to the trash bin (Recently Deleted)."""
        return self.connection.move_items_to_trash(
            self.data["drivewsid"], self.data["etag"]
        )

    def delete(self):
        """Delete an iCloud Drive item."""
        return self.connection.delete_items(self.data["drivewsid"], self.data["etag"])

    def recover(self):
        """Recovers an iCloud Drive item from trash."""
        # check to ensure item is in the trash - it should have a "restorePath" property
        if self.data.get("restorePath"):
            return self.connection.recover_items_from_trash(
                self.data["drivewsid"], self.data["etag"]
            )
        raise ValueError(f"'{self.name}' does not appear to be in the Trash.")

    def delete_forever(self):
        """Permanently deletes an iCloud Drive item from trash."""
        # check to ensure item is in the trash - it should have a "restorePath" property
        if self.data.get("restorePath"):
            return self.connection.delete_forever_from_trash(
                self.data["drivewsid"], self.data["etag"]
            )
        raise ValueError(
            f"'{self.name}' does not appear to be in the Trash. Please 'delete()' it first before "
            f"trying to 'delete_forever()'."
        )

    def get(self, name: str) -> "DriveNode":
        """Gets the node child."""
        if self.type == "file":
            raise NotADirectoryError(name)
        return [child for child in self.get_children() if child.name == name][0]

    def __getitem__(self, key: str) -> "DriveNode":
        try:
            return self.get(key)
        except IndexError as i:
            raise KeyError(f"No child named '{key}' exists") from i

    def __str__(self) -> str:
        return "{" + f"type: {self.type}, name: {self.name}" + "}"

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {str(self)}>"


def _date_to_utc(date) -> Optional[datetime]:
    if not date:
        return None
    # jump through hoops to return time in UTC rather than California time
    match: Optional[Match[str]] = search(r"^(.+?)([\+\-]\d+):(\d\d)$", date)
    if not match:
        # Already in UTC
        return datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")
    base: datetime = datetime.strptime(match.group(1), "%Y-%m-%dT%H:%M:%S")
    diff = timedelta(hours=int(match.group(2)), minutes=int(match.group(3)))
    return base - diff



"""Find my iPhone service."""

import logging
import threading
import time
from datetime import datetime, timedelta
from types import MappingProxyType
from typing import Any, Callable, Iterator, Optional

from requests import Response

from pyicloud.exceptions import (
    PyiCloudAuthRequiredException,
    PyiCloudNoDevicesException,
    PyiCloudServiceUnavailable,
)
from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession

_FMIP_CLIENT_CONTEXT_TIMEZONE: str = "US/Pacific"
_LOGGER: logging.Logger = logging.getLogger(__name__)
_MAX_REFRESH_RETRIES: int = 5


def _monitor_thread(
    interval: float, func: Callable, stop_event: threading.Event, locate: bool = True
) -> None:
    next_event: datetime = datetime.now() + timedelta(seconds=interval)
    while not stop_event.wait(timeout=interval):
        if next_event < datetime.now():
            try:
                func(locate)
            except Exception as exc:
                _LOGGER.debug("FindMyiPhone monitor thread error: %s", exc)
            next_event = datetime.now() + timedelta(seconds=interval)


class FindMyiPhoneServiceManager(BaseService):
    """The 'Find my iPhone' iCloud service

    This connects to iCloud and return phone data including the near-realtime
    latitude and longitude.
    """

    def __init__(
        self,
        service_root: str,
        token_endpoint: str,
        session: PyiCloudSession,
        params: dict[str, Any],
        with_family=False,
    ) -> None:
        """Initialize the FindMyiPhoneServiceManager."""
        super().__init__(service_root, session, params)
        self._with_family: bool = with_family

        fmip_endpoint: str = f"{service_root}/fmipservice/client/web"
        self._fmip_init_url: str = f"{fmip_endpoint}/initClient"
        self._fmip_refresh_url: str = f"{fmip_endpoint}/refreshClient"
        self._fmip_sound_url: str = f"{fmip_endpoint}/playSound"
        self._fmip_message_url: str = f"{fmip_endpoint}/sendMessage"
        self._fmip_lost_url: str = f"{fmip_endpoint}/lostDevice"
        self._fmip_erase_url: str = f"{fmip_endpoint}/remoteWipeWithUserAuth"
        self._erase_token_url: str = f"{token_endpoint}/fmipWebAuthenticate"

        self._devices: dict[str, AppleDevice] = {}
        self._devices_names: list[str] = []
        self._server_ctx: dict[str, Any] | None = None
        self._user_info: dict[str, Any] | None = None
        self._monitor: Optional[threading.Thread] = None
        self.stop_event: threading.Event = threading.Event()

        self._refresh_client_with_reauth(locate=True)

    def _refresh_client_with_reauth(
        self, retry: bool = False, locate: bool = True
    ) -> None:
        """
        Refreshes the FindMyiPhoneService endpoint with re-authentication.
        This ensures that the location data is up-to-date.
        """
        # Refresh the client (own devices first)
        self.stop_event.set()
        try:
            self._refresh_client(locate=locate)
        except PyiCloudAuthRequiredException:
            if retry is True:
                raise

            _LOGGER.debug("Re-authenticating session")
            self._server_ctx = None
            self.session.service.authenticate(force_refresh=True)
            self._refresh_client_with_reauth(retry=True, locate=locate)
            return

        # Initialize devices (including family devices if enabled)
        self._initialize_devices(locate=locate)
        self._start_monitor_thread()

    def _initialize_devices(self, locate: bool = False) -> None:
        """Initializes the devices for the FindMyiPhoneServiceManager."""

        # If family sharing is enabled, we may need to poll until all devices are ready
        # This is indicated by the deviceFetchStatus being "LOADING"
        retries: int = 0
        while (
            self._with_family
            and self._user_info
            and self._user_info.get("hasMembers", False)
        ):
            needs_refresh: bool = False
            for user in self._user_info.get("membersInfo", {}).values():
                if user.get("deviceFetchStatus") == "LOADING":
                    needs_refresh = True
                    break

            if needs_refresh:
                time.sleep(0.1)
                self._refresh_client(locate=locate)
                retries += 1
                if retries >= _MAX_REFRESH_RETRIES:
                    _LOGGER.debug("Max retries reached when fetching family devices")
                    break
            else:
                break

        if not self._devices:
            raise PyiCloudNoDevicesException()

        _LOGGER.info("Number of devices found: %d", len(self._devices))

    def _start_monitor_thread(self) -> None:
        """Starts the monitor thread for the FindMyiPhoneServiceManager."""
        if not self.is_alive:
            self._monitor = threading.Thread(
                target=_monitor_thread,
                kwargs={
                    "func": self._refresh_client,
                    "interval": 1.0,
                    "stop_event": self.stop_event,
                },
                daemon=True,
            )
            self.stop_event.clear()
            self._monitor.start()

    def _refresh_client(self, locate: bool = False) -> None:
        """
        Refreshes the FindMyiPhoneService endpoint, this ensures that the location data
        is up-to-date.
        """
        req_json: dict[str, Any] = {
            "clientContext": {
                "appName": "iCloud Find (Web)",
                "appVersion": "2.0",
                "apiVersion": "3.0",
                "deviceListVersion": 1,
                "fmly": self._with_family,
                "timezone": _FMIP_CLIENT_CONTEXT_TIMEZONE,
                "inactiveTime": 0,
            },
        }

        if self._server_ctx:
            req_json["serverContext"] = self._server_ctx
            if locate:
                req_json["isUpdatingAllLocations"] = True
                req_json["clientContext"].update(
                    {
                        "shouldLocate": True,
                        "selectedDevice": "all",
                    }
                )

        req: Response = self.session.post(
            url=self._fmip_refresh_url if self._server_ctx else self._fmip_init_url,
            params=self.params,
            json=req_json,
        )
        resp: dict[str, Any] = req.json()

        self._server_ctx = resp.get("serverContext")
        if self._server_ctx and "theftLoss" in self._server_ctx:
            self._server_ctx["theftLoss"] = None

        self._user_info = resp.get("userInfo")

        if "content" not in resp:
            _LOGGER.debug("FMIP returned 0 devices")
            return

        _LOGGER.debug("FMIP returned %d devices", len(resp["content"]))
        for device_info in resp["content"]:
            device_id: str = device_info["id"]
            if device_id not in self._devices:
                self._devices[device_id] = AppleDevice(
                    device_info,
                    self.params,
                    manager=self,
                    sound_url=self._fmip_sound_url,
                    lost_url=self._fmip_lost_url,
                    message_url=self._fmip_message_url,
                    erase_url=self._fmip_erase_url,
                    erase_token_url=self._erase_token_url,
                )
            else:
                self._devices[device_id].update(device_info)

        self._devices_names = list(self._devices.keys())

    def refresh(self, locate: bool = False) -> None:
        """Public method to refresh the FindMyiPhoneService endpoint."""
        self._refresh_client_with_reauth(locate=locate)

    def __getitem__(self, key: str | int) -> "AppleDevice":
        """Gets a device by name or index."""
        if not self.is_alive:
            self._refresh_client_with_reauth(locate=True)

        if isinstance(key, int):
            key = self._devices_names[key]
        return self._devices[key]

    def __str__(self) -> str:
        """String representation of the devices."""
        return f"{self._devices}"

    def __repr__(self) -> str:
        """Representation of the device."""
        return f"{self}"

    def __iter__(self) -> Iterator["AppleDevice"]:
        """Iterates over the devices."""

        if not self.is_alive:
            self._refresh_client_with_reauth(locate=True)
        return iter(self._devices.values())

    def __len__(self) -> int:
        """Returns the number of devices."""
        if not self.is_alive:
            self._refresh_client_with_reauth(locate=True)
        return len(self._devices)

    @property
    def is_alive(self) -> bool:
        """Indicates if the service is alive."""
        return self._monitor is not None and self._monitor.is_alive()

    @property
    def devices(self) -> "MappingProxyType[str, AppleDevice]":
        """Returns the devices."""
        return MappingProxyType(self._devices)

    @property
    def user_info(self) -> Optional[MappingProxyType[str, Any]]:
        """Returns the user info."""
        return MappingProxyType(self._user_info) if self._user_info else None


class AppleDevice:
    """Apple device."""

    def __init__(
        self,
        content: dict[str, Any],
        params: dict[str, Any],
        manager: FindMyiPhoneServiceManager,
        sound_url: str,
        lost_url: str,
        erase_url: str,
        erase_token_url: str,
        message_url: str,
    ) -> None:
        """Initialize the Apple device."""
        self._content: dict[str, Any] = content
        self._manager: FindMyiPhoneServiceManager = manager
        self._params: dict[str, Any] = params

        self._sound_url: str = sound_url
        self._lost_url: str = lost_url
        self._erase_url: str = erase_url
        self._erase_token_url: str = erase_token_url
        self._message_url: str = message_url

    @property
    def session(self) -> PyiCloudSession:
        """Gets the session."""
        return self._manager.session

    def update(self, data) -> None:
        """Updates the device data."""
        self._content = data

    @property
    def location(self) -> Optional[dict[str, Any]]:
        """Updates the device location."""
        if self.location_available is False:
            return None
        return self._content["location"]

    def status(self, additional: Optional[list[str]] = None) -> dict[str, Any]:
        """Returns status information for device.

        This returns only a subset of possible properties.
        """
        fields: list[str] = [
            "batteryLevel",
            "deviceDisplayName",
            "deviceStatus",
            "name",
        ]
        if additional is not None:
            fields += additional

        properties: dict[str, Any] = {}
        for field in fields:
            properties[field] = self._content.get(field)
        return properties

    def play_sound(self, subject="Find My iPhone Alert") -> None:
        """Send a request to the device to play a sound.

        It's possible to pass a custom message by changing the `subject`.
        """
        if self.sound_available is False:
            raise PyiCloudServiceUnavailable("Sound is not available for this device")
        data: dict[str, Any] = {
            "device": self._content["id"],
            "subject": subject,
            "clientContext": {"fmly": True},
        }

        self.session.post(self._sound_url, params=self._params, json=data)

    def display_message(
        self,
        subject="Find My iPhone Alert",
        message="This is a note",
        sounds=False,
        vibrate=False,
        strobe=False,
    ) -> None:
        """Send a request to the device to a display a message.

        It's possible to pass a custom message by changing the `subject`.
        """
        if self.messaging_available is False:
            raise PyiCloudServiceUnavailable(
                "Messaging is not available for this device"
            )

        data: dict[str, Any] = {
            "device": self._content["id"],
            "subject": subject,
            "userText": True,
            "text": message,
            "sound": sounds,
            "vibrate": vibrate,
            "strobe": strobe,
        }

        self.session.post(self._message_url, params=self._params, json=data)

    def lost_device(
        self,
        number: str,
        text: str = "This device has been lost. Please call me.",
        newpasscode: str = "",
    ) -> None:
        """Send a request to the device to trigger 'lost mode'.

        The device will show the message in `text`, and if a number has
        been passed, then the person holding the device can call
        the number without entering the passcode.
        """
        if self.lost_mode_available is False:
            raise PyiCloudServiceUnavailable(
                "Lost mode is not available for this device"
            )

        data: dict[str, Any] = {
            "text": text,
            "userText": True,
            "ownerNbr": number,
            "lostModeEnabled": True,
            "trackingEnabled": True,
            "device": self._content["id"],
            "passcode": newpasscode,
        }

        self.session.post(self._lost_url, params=self._params, json=data)

    def _get_erase_token(self) -> str:
        """Get the erase token for the Find My iPhone service."""
        data: dict[str, Any] = {
            "dsWebAuthToken": self.session.data.get("session_token"),
        }

        data = self.session.post(url=self._erase_token_url, json=data).json()
        if "tokens" not in data or "mmeFMIPWebEraseDeviceToken" not in data["tokens"]:
            raise PyiCloudServiceUnavailable("Find My iPhone erase token not available")
        return data["tokens"]["mmeFMIPWebEraseDeviceToken"]

    def erase_device(
        self,
        text: str = "This device has been lost. Please call me.",
        newpasscode: str = "",
    ) -> None:
        """Send a request to the device to start a remote erase."""
        if self.erase_available is False:
            raise PyiCloudServiceUnavailable("Erase is not available for this device")

        data: dict[str, Any] = {
            "authToken": self._get_erase_token(),
            "text": text,
            "device": self._content["id"],
            "passcode": newpasscode,
        }

        self.session.post(self._erase_url, params=self._params, json=data)

    @property
    def data(self) -> dict[str, Any]:
        """Gets the device data."""
        if not self._manager.is_alive:
            self._manager.refresh()

        return self._content

    def __getitem__(self, key) -> Any:
        """Gets an attribute of the device data."""
        if not self._manager.is_alive:
            self._manager.refresh()

        return self._content[key]

    def __getattr__(self, attr) -> Any:
        """Gets an attribute of the device data."""
        if not self._manager.is_alive:
            self._manager.refresh()

        if attr in self._content:
            return self._content[attr]
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{attr}'"
        )

    @property
    def name(self) -> str:
        """Gets the device name."""
        return self.data.get("name", "")

    @property
    def model(self) -> str:
        """Gets the device model."""
        return self.data.get("deviceModel", "")

    @property
    def model_name(self) -> str:
        """Gets the device model name."""
        return self.data.get("deviceDisplayName", "")

    @property
    def device_type(self) -> str:
        """Gets the device type."""
        return self.data.get("deviceClass", "")

    @property
    def lost_mode_available(self) -> bool:
        """Indicates if lost mode is available for the device."""
        return self.data.get("lostModeCapable", False)

    @property
    def messaging_available(self) -> bool:
        """Indicates if messaging is available for the device."""
        return self.data.get("features", {}).get("MSG", False)

    @property
    def sound_available(self) -> bool:
        """Indicates if sound is available for the device."""
        return self.data.get("features", {}).get("SND", False)

    @property
    def erase_available(self) -> bool:
        """Indicates if erase is available for the device."""
        return self.data.get("features", {}).get("WIP", False)

    @property
    def location_available(self) -> bool:
        """Indicates if location is available for the device."""
        return (
            self.data.get("features", {}).get("LOC", False)
            and "location" in self._content
            and self._content["location"] is not None
        )

    def __str__(self) -> str:
        """String representation of the device."""
        return f"{self.model_name}: {self.name}"

    def __repr__(self) -> str:
        """Representation of the device."""
        return f"<AppleDevice({self})>"



"""Library base file."""

import base64
import getpass
import json
import logging
import time
from os import environ, mkdir, path
from tempfile import gettempdir
from typing import Any, Dict, List, Optional
from uuid import uuid1

import srp
from fido2.client import DefaultClientDataCollector, Fido2Client
from fido2.hid import CtapHidDevice
from fido2.webauthn import (
    AuthenticationResponse,
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialRequestOptions,
    PublicKeyCredentialType,
    UserVerificationRequirement,
)
from requests import HTTPError
from requests.models import Response

from pyicloud.const import ACCOUNT_NAME, CONTENT_TYPE_JSON, CONTENT_TYPE_TEXT
from pyicloud.exceptions import (
    PyiCloud2FARequiredException,
    PyiCloudAcceptTermsException,
    PyiCloudAPIResponseException,
    PyiCloudFailedLoginException,
    PyiCloudNoTrustedNumberAvailable,
    PyiCloudPasswordException,
    PyiCloudServiceNotActivatedException,
    PyiCloudServiceUnavailable,
)
from pyicloud.services import (
    AccountService,
    AppleDevice,
    CalendarService,
    ContactsService,
    DriveService,
    FindMyiPhoneServiceManager,
    HideMyEmailService,
    PhotosService,
    RemindersService,
    UbiquityService,
)
from pyicloud.session import PyiCloudSession
from pyicloud.srp_password import SrpPassword, SrpProtocolType
from pyicloud.utils import (
    b64_encode,
    b64url_decode,
    get_password_from_keyring,
)

LOGGER: logging.Logger = logging.getLogger(__name__)
PCS_SLEEP_TIME: int = 5
PCS_MAX_RETRIES: int = 10

_HEADERS: dict[str, str] = {
    "User-Agent": (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
        "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.3.1 Safari/605.1.15"
    ),
}

_AUTH_HEADERS_JSON: dict[str, str] = {
    "Accept": f"{CONTENT_TYPE_JSON}, text/javascript",
    "Content-Type": CONTENT_TYPE_JSON,
    "X-Apple-OAuth-Client-Id": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
    "X-Apple-OAuth-Client-Type": "firstPartyAuth",
    "X-Apple-OAuth-Redirect-URI": "https://www.icloud.com",
    "X-Apple-OAuth-Require-Grant-Code": "true",
    "X-Apple-OAuth-Response-Mode": "web_message",
    "X-Apple-OAuth-Response-Type": "code",
    "X-Apple-OAuth-State": "",
    "X-Apple-Widget-Key": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
    "X-Apple-FD-Client-Info": json.dumps(
        {
            "U": _HEADERS["User-Agent"],
            "L": "en-US",
            "Z": "GMT+00:00",
            "V": "1.1",
            "F": "",
        },
        separators=(",", ":"),
    ),
}

_PARAMS: dict[str, str] = {
    "clientBuildNumber": "2534Project66",
    "clientMasteringNumber": "2534B22",
}


class PyiCloudService:
    """
    A base authentication class for the iCloud service. Handles the
    authentication required to access iCloud services.

    Usage:
        from pyicloud import PyiCloudService
        pyicloud = PyiCloudService('username@apple.com', 'password')
        pyicloud.iphone.location()
    """

    def _setup_endpoints(self) -> None:
        """Set up the endpoints for the service."""
        # If the country or region setting of your Apple ID is China mainland.
        # See https://support.apple.com/en-us/HT208351
        icloud_china: str = ".cn" if self._is_china_mainland else ""
        self._idmsa_endpoint: str = f"https://idmsa.apple.com{icloud_china}"
        self._auth_endpoint: str = f"{self._idmsa_endpoint}/appleauth/auth"
        self._home_endpoint: str = f"https://www.icloud.com{icloud_china}"
        self._setup_endpoint: str = f"https://setup.icloud.com{icloud_china}/setup/ws/1"

    def _setup_cookie_directory(self, cookie_directory: Optional[str]) -> str:
        """Set up the cookie directory for the service."""
        _cookie_directory: str = ""
        if cookie_directory:
            _cookie_directory = path.expanduser(path.normpath(cookie_directory))
            if not path.exists(_cookie_directory):
                mkdir(_cookie_directory, 0o700)
        else:
            topdir: str = path.join(gettempdir(), "pyicloud")
            _cookie_directory = path.join(topdir, getpass.getuser())
            if not path.exists(topdir):
                mkdir(topdir, 0o777)
            if not path.exists(_cookie_directory):
                mkdir(_cookie_directory, 0o700)

        return _cookie_directory

    def __init__(
        self,
        apple_id: str,
        password: Optional[str] = None,
        cookie_directory: Optional[str] = None,
        verify: bool = True,
        client_id: Optional[str] = None,
        with_family: bool = True,
        china_mainland: bool = False,
        accept_terms: bool = False,
    ) -> None:
        self._is_china_mainland: bool = (
            china_mainland or environ.get("icloud_china", "0") == "1"
        )
        self._setup_endpoints()

        self._password_raw: Optional[str] = password

        self._apple_id: str = apple_id
        self._accept_terms: bool = accept_terms

        if self._password_raw is None:
            self._password_raw = get_password_from_keyring(apple_id)

        self.data: dict[str, Any] = {}
        self._auth_data: dict[str, Any] = {}

        self.params: dict[str, Any] = {}
        self._client_id: str = client_id or str(uuid1()).lower()
        self._with_family: bool = with_family

        _cookie_directory: str = self._setup_cookie_directory(cookie_directory)
        _headers: dict[str, str] = _HEADERS.copy()
        _headers.update(
            {
                "Origin": self._home_endpoint,
                "Referer": f"{self._home_endpoint}/",
            }
        )

        self._session: PyiCloudSession = PyiCloudSession(
            self,
            verify=verify,
            headers=_headers,
            client_id=self._client_id,
            cookie_directory=_cookie_directory,
        )

        self._client_id = self.session.data.get("client_id", self._client_id)

        _params: dict[str, str] = _PARAMS.copy()
        _params.update(
            {
                "clientId": self._client_id,
            }
        )
        self.params = _params

        self._webservices: Optional[dict[str, dict[str, Any]]] = None

        self._account: Optional[AccountService] = None
        self._calendar: Optional[CalendarService] = None
        self._contacts: Optional[ContactsService] = None
        self._devices: Optional[FindMyiPhoneServiceManager] = None
        self._drive: Optional[DriveService] = None
        self._files: Optional[UbiquityService] = None
        self._hidemyemail: Optional[HideMyEmailService] = None
        self._photos: Optional[PhotosService] = None
        self._reminders: Optional[RemindersService] = None

        self._requires_mfa: bool = False

        self.authenticate()

    def authenticate(
        self, force_refresh: bool = False, service: Optional[str] = None
    ) -> None:
        """
        Handles authentication, and persists cookies so that
        subsequent logins will not cause additional e-mails from Apple.
        """

        login_successful = False
        if self.session.data.get("session_token") and not force_refresh:
            try:
                self.data = self._validate_token()
                login_successful = True
            except PyiCloudAPIResponseException:
                LOGGER.debug("Invalid authentication token, will log in from scratch.")

        if (
            not login_successful
            and service is not None
            and self.data.get("apps")
            and service in self.data["apps"]
        ):
            app: dict[str, Any] = self.data["apps"][service]
            if "canLaunchWithOneFactor" in app and app["canLaunchWithOneFactor"]:
                LOGGER.debug("Authenticating as %s for %s", self.account_name, service)
                try:
                    self._authenticate_with_credentials_service(service)
                    login_successful = True
                except PyiCloudFailedLoginException:
                    LOGGER.debug(
                        "Could not log into service. Attempting brand new login."
                    )

        if not login_successful:
            try:
                self._authenticate()
                LOGGER.debug("Authentication completed successfully")
            except PyiCloud2FARequiredException:
                self._requires_mfa = True
                LOGGER.debug("2FA is required")

        self._update_state()

    def _handle_accept_terms(self, login_data: dict) -> None:
        """Handle accepting updated terms of service."""
        if self.data.get("termsUpdateNeeded"):
            if not self._accept_terms:
                raise PyiCloudAcceptTermsException(
                    "You must accept the updated terms of service to continue. "
                    "Set --accept-terms to accept them."
                )
            resp: Response = self.session.get(
                f"{self._setup_endpoint}/getTerms",
                params=self.params,
                json={
                    "locale": self.data.get("dsInfo", {}).get("languageCode", "en_US")
                },
            )
            resp.raise_for_status()
            terms_info: dict[str, Any] = resp.json()
            version: int | None = terms_info.get("iCloudTerms", {}).get("version")
            if version is None:
                raise PyiCloudAcceptTermsException("Could not get terms version")
            resp = self.session.get(
                f"{self._setup_endpoint}/repairDone",
                params=self.params,
                json={"acceptedICloudTerms": version},
            )
            resp.raise_for_status()

            resp = self.session.post(
                f"{self._setup_endpoint}/accountLogin", json=login_data
            )
            resp.raise_for_status()

            self.data = resp.json()

    def _update_state(self) -> None:
        """Update the state of the service."""
        if (
            "dsInfo" in self.data
            and isinstance(self.data["dsInfo"], dict)
            and "dsid" in self.data["dsInfo"]
        ):
            self.params.update({"dsid": self.data["dsInfo"]["dsid"]})

        if "webservices" in self.data:
            self._webservices = self.data["webservices"]

    def _authenticate(self) -> None:
        LOGGER.debug("Authenticating as %s", self.account_name)

        try:
            self._authenticate_with_token()
        except (PyiCloudFailedLoginException, PyiCloud2FARequiredException):
            self._srp_authentication()
            self._authenticate_with_token()

    def _srp_authentication(self) -> None:
        """SRP authentication."""
        if self._password_raw is None:
            raise PyiCloudFailedLoginException("No password set")

        auth_headers = self._get_auth_headers()

        response: Response = self.session.get(
            f"{self._auth_endpoint}/authorize/signin",
            params={
                "frame_id": auth_headers["X-Apple-OAuth-State"],
                "skVersion": "7",
                "iframeid": auth_headers["X-Apple-OAuth-State"],
                "client_id": auth_headers["X-Apple-Widget-Key"],
                "response_type": auth_headers["X-Apple-OAuth-Response-Type"],
                "redirect_uri": auth_headers["X-Apple-OAuth-Redirect-URI"],
                "response_mode": auth_headers["X-Apple-OAuth-Response-Mode"],
                "state": auth_headers["X-Apple-OAuth-State"],
                "authVersion": "latest",
            },
        )
        response.raise_for_status()

        srp_password: SrpPassword = SrpPassword(self._password_raw)
        srp.rfc5054_enable()
        srp.no_username_in_x()
        try:
            usr = srp.User(
                self.account_name,
                srp_password,
                hash_alg=srp.SHA256,
                ng_type=srp.NG_2048,
            )
            uname, A = usr.start_authentication()  # pylint: disable=invalid-name
            data: dict[str, Any] = {
                "a": b64_encode(A),
                ACCOUNT_NAME: uname,
                "protocols": [protocol.value for protocol in SrpProtocolType],
            }

            response: Response = self.session.post(
                f"{self._auth_endpoint}/signin/init",
                json=data,
                headers=self._get_auth_headers(),
            )
            response.raise_for_status()
        except (
            PyiCloudAPIResponseException,
            HTTPError,
            PyiCloudPasswordException,
        ) as error:
            msg = "Failed to initiate srp authentication."
            raise PyiCloudFailedLoginException(msg, error) from error

        body: dict[str, Any] = response.json()
        salt: bytes = base64.b64decode(body["salt"])
        b: bytes = base64.b64decode(body["b"])
        c: Any = body["c"]
        iterations: int = body["iteration"]
        protocol: SrpProtocolType = SrpProtocolType(body["protocol"])
        key_length: int = 32

        srp_password.set_encrypt_info(salt, iterations, key_length, protocol)

        m1: None | Any = usr.process_challenge(salt, b)
        m2: None | bytes = usr.H_AMK
        if m1 and m2:
            data = {
                ACCOUNT_NAME: uname,
                "c": c,
                "m1": b64_encode(m1),
                "m2": b64_encode(m2),
                "rememberMe": True,
                "trustTokens": [],
            }
        if self.session.data.get("trust_token"):
            data["trustTokens"] = [self.session.data.get("trust_token")]

        try:
            self.session.post(
                f"{self._auth_endpoint}/signin/complete",
                params={
                    "isRememberMeEnabled": "true",
                },
                json=data,
                headers=self._get_auth_headers(),
            )
        except PyiCloud2FARequiredException:
            LOGGER.debug("2FA required to complete authentication.")
            self._auth_data = self._get_mfa_auth_options()
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg) from error

    def _authenticate_with_token(self) -> None:
        """Authenticate using session token."""
        if not self.session.data.get("session_token"):
            raise PyiCloudFailedLoginException("No session token available")

        try:
            login_data: dict[str, Any] = {
                "accountCountryCode": self.session.data.get("account_country"),
                "dsWebAuthToken": self.session.data.get("session_token"),
                "extended_login": True,
                "trustToken": self.session.data.get("trust_token", ""),
            }

            resp: Response = self.session.post(
                f"{self._setup_endpoint}/accountLogin", json=login_data
            )
            resp.raise_for_status()

            self.data = resp.json()

            self._handle_accept_terms(login_data)

            if not self.is_trusted_session:
                raise PyiCloud2FARequiredException(self.account_name, resp)
        except (PyiCloudAPIResponseException, HTTPError) as error:
            msg = "Invalid authentication token."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _authenticate_with_credentials_service(self, service: Optional[str]) -> None:
        """Authenticate to a specific service using credentials."""
        login_data: dict[str, Any] = {
            "appName": service,
            "apple_id": self.account_name,
            "password": self._password_raw,
        }

        try:
            self.session.post(f"{self._setup_endpoint}/accountLogin", json=login_data)

            self._handle_accept_terms(login_data)

            self.data = self._validate_token()
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _validate_token(self) -> Any:
        """Checks if the current access token is still valid."""
        LOGGER.debug("Checking session token validity")
        if not self.session.cookies.get("X-APPLE-WEBAUTH-TOKEN"):
            raise PyiCloudAPIResponseException(
                "Missing X-APPLE-WEBAUTH-TOKEN cookie", None
            )
        try:
            req: Response = self.session.post(
                f"{self._setup_endpoint}/validate", data="null"
            )
            LOGGER.debug("Session token is still valid")
            return req.json()
        except PyiCloudAPIResponseException:
            LOGGER.debug("Invalid authentication token")
            raise

    def _get_auth_headers(
        self, overrides: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        headers: dict[str, Any] = _AUTH_HEADERS_JSON.copy()
        headers.update(
            {
                "Referer": self._idmsa_endpoint,
                "X-Apple-OAuth-State": self._client_id,
                "X-Apple-Frame-Id": self._client_id,
            }
        )

        if self.session.data.get("scnt"):
            headers["scnt"] = self.session.data["scnt"]

        if self.session.data.get("session_id"):
            headers["X-Apple-ID-Session-Id"] = self.session.data["session_id"]
        if self.session.data.get("auth_attributes"):
            headers["X-Apple-Auth-Attributes"] = self.session.data["auth_attributes"]

        if overrides:
            headers.update(overrides)

        return headers

    @property
    def session(self) -> PyiCloudSession:
        """Return the session."""
        return self._session

    def _is_mfa_required(self) -> bool:
        return (
            self.data.get("hsaChallengeRequired", False)
            or not self.is_trusted_session
            or self._requires_mfa
        )

    @property
    def requires_2sa(self) -> bool:
        """Returns True if two-step authentication is required."""
        return (
            self._is_mfa_required()
            and self.data.get("dsInfo", {}).get("hsaVersion", 0) >= 1
        )

    @property
    def requires_2fa(self) -> bool:
        """Returns True if two-factor authentication is required."""
        return (
            self._is_mfa_required()
            and self.data.get("dsInfo", {}).get("hsaVersion", 0) == 2
        )

    @property
    def is_trusted_session(self) -> bool:
        """Returns True if the session is trusted."""
        return self.data.get("hsaTrustedBrowser", False)

    @property
    def trusted_devices(self) -> list[dict[str, Any]]:
        """Returns devices trusted for two-step authentication."""
        request: Response = self.session.get(
            f"{self._setup_endpoint}/listDevices", params=self.params
        )
        return request.json().get("devices")

    def send_verification_code(self, device: dict[str, Any]) -> bool:
        """Requests that a verification code is sent to the given device."""
        request: Response = self.session.post(
            f"{self._setup_endpoint}/sendVerificationCode",
            params=self.params,
            json=device,
        )
        return request.json().get("success", False)

    def validate_verification_code(self, device: dict[str, Any], code: str) -> bool:
        """Verifies a verification code received on a trusted device."""

        device.update({"verificationCode": code, "trustBrowser": True})

        try:
            self.session.post(
                f"{self._setup_endpoint}/validateVerificationCode",
                params=self.params,
                json=device,
            )
        except PyiCloudAPIResponseException as error:
            if error.code == -21669:
                # Wrong verification code
                return False
            raise

        self.trust_session()

        return not self.requires_2sa

    def _get_mfa_auth_options(self) -> Dict:
        """Retrieve auth request options for assertion."""
        headers = self._get_auth_headers({"Accept": CONTENT_TYPE_JSON})

        return self.session.get(self._auth_endpoint, headers=headers).json()

    @property
    def security_key_names(self) -> Optional[List[str]]:
        """Security key names which can be used for the WebAuthn assertion."""
        return self._auth_data.get("keyNames")

    def _submit_webauthn_assertion_response(self, data: Dict) -> None:
        """Submit the WebAuthn assertion response for authentication."""
        headers = self._get_auth_headers({"Accept": CONTENT_TYPE_JSON})

        self.session.post(
            f"{self._auth_endpoint}/verify/security/key", json=data, headers=headers
        )

    @property
    def fido2_devices(self) -> List[CtapHidDevice]:
        """List the available FIDO2 devices."""
        return list(CtapHidDevice.list_devices())

    def confirm_security_key(self, device: Optional[CtapHidDevice] = None) -> None:
        """Conduct the WebAuthn assertion ceremony with user's FIDO2 device."""
        fsa: dict[str, Any] = self._auth_data.get("fsaChallenge", {})
        try:
            challenge = fsa["challenge"]
            allowed_credentials = fsa["keyHandles"]
            rp_id = fsa["rpId"]
        except KeyError as error:
            raise PyiCloudAPIResponseException(
                "Missing WebAuthn challenge data"
            ) from error

        if not device:
            devices: List[CtapHidDevice] = list(CtapHidDevice.list_devices())

            if not devices:
                raise RuntimeError("No FIDO2 devices found")

            device = devices[0]

        client = Fido2Client(
            device,
            client_data_collector=DefaultClientDataCollector("https://apple.com"),
        )
        credentials: List[PublicKeyCredentialDescriptor] = [
            PublicKeyCredentialDescriptor(
                id=b64url_decode(cred_id), type=PublicKeyCredentialType("public-key")
            )
            for cred_id in allowed_credentials
        ]
        assertion_options = PublicKeyCredentialRequestOptions(
            challenge=b64url_decode(challenge),
            rp_id=rp_id,
            allow_credentials=credentials,
            user_verification=UserVerificationRequirement("discouraged"),
        )
        result: AuthenticationResponse = client.get_assertion(
            assertion_options
        ).get_response(0)

        self._submit_webauthn_assertion_response(
            {
                "challenge": challenge,
                "clientData": b64_encode(result.response.client_data),
                "signatureData": b64_encode(result.response.signature),
                "authenticatorData": b64_encode(result.response.authenticator_data),
                "userHandle": b64_encode(result.response.user_handle)
                if result.response.user_handle
                else None,
                "credentialID": b64_encode(result.raw_id),
                "rpId": rp_id,
            }
        )

        self.trust_session()

    def _check_pcs_consent(self) -> dict[str, Any]:
        """Check if the user has consented to PCS access."""
        LOGGER.debug("Querying web access state")
        resp = self.session.post(
            f"{self._setup_endpoint}/requestWebAccessState", params=self.params
        ).json()

        return resp

    def _send_pcs_request(
        self, app_name: str, derived_from_user_action: bool
    ) -> dict[str, Any]:
        """Send a request to the PCS endpoint to check the status of PCS access."""
        LOGGER.debug("Querying PCS status")

        return self.session.post(
            f"{self._setup_endpoint}/requestPCS",
            json={
                "appName": app_name,
                "derivedFromUserAction": derived_from_user_action,
            },
            params=self.params,
        ).json()

    def _request_pcs_for_service(self, app_name: str) -> None:
        """Request PCS access for a specific service."""
        _check_pcs_resp: dict[str, Any] = self._check_pcs_consent()

        if not _check_pcs_resp.get("isICDRSDisabled", False):
            LOGGER.warning("ICDRS is not disabled")
            return

        if not _check_pcs_resp.get("isDeviceConsentedForPCS", True):
            LOGGER.debug("Requesting PCS consent")

            resp = self.session.post(
                f"{self._setup_endpoint}/enableDeviceConsentForPCS", params=self.params
            ).json()

            if not resp.get("isDeviceConsentNotificationSent"):
                raise PyiCloudAPIResponseException("Unable to request PCS access!")

        LOGGER.debug("Waiting for PCS consent")
        for _ in range(PCS_MAX_RETRIES):
            if _check_pcs_resp.get("isDeviceConsentedForPCS", True):
                LOGGER.debug("PCS consent granted")
                break

            LOGGER.debug("PCS consent not granted yet, waiting...")
            time.sleep(PCS_SLEEP_TIME)
            _check_pcs_resp = self._check_pcs_consent()

        for attempt in range(PCS_MAX_RETRIES):
            resp: dict[str, Any] = self._send_pcs_request(
                app_name,
                derived_from_user_action=attempt == 0,
            )

            if resp["status"] == "success":
                LOGGER.debug("PCS access was granted")
                return

            if resp["message"] in (
                "Requested the device to upload cookies.",
                "Cookies not available yet on server.",
            ):
                LOGGER.debug("PCS access couldn't be obtained: %s", resp["message"])
                time.sleep(PCS_SLEEP_TIME)
            else:
                LOGGER.error("Unknown PCS state: %s", resp["message"])
                raise PyiCloudAPIResponseException("Unable to request PCS access!")

    def validate_2fa_code(self, code: str) -> bool:
        """Verifies a verification code received via Apple's 2FA system (HSA2)."""
        try:
            if self._auth_data.get("mode") == "sms":
                self._validate_sms_code(code)
            else:
                data: dict[str, Any] = {"securityCode": {"code": code}}
                headers: dict[str, Any] = self._get_auth_headers(
                    {"Accept": CONTENT_TYPE_JSON}
                )
                self.session.post(
                    f"{self._auth_endpoint}/verify/trusteddevice/securitycode",
                    json=data,
                    headers=headers,
                )
        except PyiCloudAPIResponseException:
            # Wrong verification code
            LOGGER.error("Code verification failed.")
            return False

        LOGGER.debug("Code verification successful.")

        self.trust_session()
        return not self.requires_2sa

    def _validate_sms_code(self, code: str) -> None:
        """Verifies a verification code received via Apple's SMS system."""
        trusted_phone_number: dict[str, Any] | None = self._auth_data.get(
            "trustedPhoneNumber"
        )
        if not trusted_phone_number:
            raise PyiCloudNoTrustedNumberAvailable()

        device_id: int | None = trusted_phone_number.get("id")
        non_fteu: bool | None = trusted_phone_number.get("nonFTEU")
        mode: str | None = trusted_phone_number.get("pushMode")

        data: dict[str, Any] = {
            "phoneNumber": {"id": device_id, "nonFTEU": non_fteu},
            "securityCode": {"code": code},
            "mode": mode,
        }
        headers: dict[str, Any] = self._get_auth_headers(
            {"Accept": f"{CONTENT_TYPE_JSON}, {CONTENT_TYPE_TEXT}"}
        )

        self.session.post(
            f"{self._auth_endpoint}/verify/phone/securitycode",
            json=data,
            headers=headers,
        )

    def trust_session(self) -> bool:
        """Request session trust to avoid user log in going forward."""
        self._requires_mfa = False

        headers: dict[str, Any] = self._get_auth_headers()

        try:
            self.session.get(
                f"{self._auth_endpoint}/2sv/trust",
                headers=headers,
            )
            self._authenticate_with_token()
            LOGGER.debug("Session trust successful.")
            return True
        except (PyiCloudAPIResponseException, PyiCloud2FARequiredException):
            LOGGER.error("Session trust failed.")
            return False

    def get_webservice_url(self, ws_key: str) -> str:
        """Get webservice URL, raise an exception if not exists."""
        if self._webservices is None or self._webservices.get(ws_key) is None:
            raise PyiCloudServiceNotActivatedException(
                f"Webservice not available: {ws_key}"
            )

        return self._webservices[ws_key]["url"]

    @property
    def devices(self) -> FindMyiPhoneServiceManager:
        """Returns all devices."""
        if not self._devices:
            try:
                service_root: str = self.get_webservice_url("findme")
                self._devices = FindMyiPhoneServiceManager(
                    service_root=service_root,
                    token_endpoint=self._setup_endpoint,
                    session=self.session,
                    params=self.params,
                    with_family=self._with_family,
                )
            except PyiCloudServiceNotActivatedException as error:
                raise PyiCloudServiceUnavailable(
                    "Find My iPhone service not available"
                ) from error
        return self._devices

    @property
    def hidemyemail(self) -> HideMyEmailService:
        """Gets the 'HME' service."""
        if not self._hidemyemail:
            service_root: str = self.get_webservice_url("premiummailsettings")
            try:
                self._hidemyemail = HideMyEmailService(
                    service_root=service_root,
                    session=self.session,
                    params=self.params,
                )
            except PyiCloudAPIResponseException as error:
                raise PyiCloudServiceUnavailable(
                    "Hide My Email service not available"
                ) from error
        return self._hidemyemail

    @property
    def iphone(self) -> AppleDevice:
        """Returns the iPhone."""
        return self.devices[0]

    @property
    def account(self) -> AccountService:
        """Gets the 'Account' service."""
        if not self._account:
            service_root: str = self.get_webservice_url("account")
            try:
                self._account = AccountService(
                    service_root=service_root,
                    session=self.session,
                    china_mainland=self._is_china_mainland,
                    params=self.params,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Account service not available"
                ) from error
        return self._account

    @property
    def files(self) -> UbiquityService:
        """Gets the 'File' service."""
        if not self._files:
            service_root: str = self.get_webservice_url("ubiquity")
            try:
                self._files = UbiquityService(
                    service_root=service_root,
                    session=self.session,
                    params=self.params,
                )
            except (PyiCloudAPIResponseException,) as error:
                if "Account migrated" == error.reason:
                    raise PyiCloudServiceUnavailable(
                        "Files service not available use `api.drive` instead"
                    ) from error
                raise PyiCloudServiceUnavailable(
                    "Files service not available"
                ) from error
        return self._files

    @property
    def photos(self) -> PhotosService:
        """Gets the 'Photo' service."""
        self._request_pcs_for_service("photos")

        if not self._photos:
            service_root: str = self.get_webservice_url("ckdatabasews")
            upload_url: str = self.get_webservice_url("uploadimagews")
            shared_streams_url: str = self.get_webservice_url("sharedstreams")
            self.params["dsid"] = self.data["dsInfo"]["dsid"]

            try:
                self._photos = PhotosService(
                    service_root=service_root,
                    session=self.session,
                    params=self.params,
                    upload_url=upload_url,
                    shared_streams_url=shared_streams_url,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Photos service not available"
                ) from error
        return self._photos

    @property
    def calendar(self) -> CalendarService:
        """Gets the 'Calendar' service."""
        if not self._calendar:
            service_root: str = self.get_webservice_url("calendar")
            try:
                self._calendar = CalendarService(
                    service_root=service_root, session=self.session, params=self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Calendar service not available"
                ) from error
        return self._calendar

    @property
    def contacts(self) -> ContactsService:
        """Gets the 'Contacts' service."""
        if not self._contacts:
            service_root: str = self.get_webservice_url("contacts")
            try:
                self._contacts = ContactsService(
                    service_root=service_root, session=self.session, params=self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Contacts service not available"
                ) from error
        return self._contacts

    @property
    def reminders(self) -> RemindersService:
        """Gets the 'Reminders' service."""
        if not self._reminders:
            service_root: str = self.get_webservice_url("reminders")
            try:
                self._reminders = RemindersService(
                    service_root=service_root, session=self.session, params=self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Reminders service not available"
                ) from error
        return self._reminders

    @property
    def drive(self) -> DriveService:
        """Gets the 'Drive' service."""
        self._request_pcs_for_service("iclouddrive")

        if not self._drive:
            try:
                self._drive = DriveService(
                    service_root=self.get_webservice_url("drivews"),
                    document_root=self.get_webservice_url("docws"),
                    session=self.session,
                    params=self.params,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Drive service not available"
                ) from error
        return self._drive

    @property
    def account_name(self) -> str:
        """Retrieves the account name associated with the Apple ID."""

        return self._apple_id

    def __str__(self) -> str:
        return f"iCloud API: {self.account_name}"

    def __repr__(self) -> str:
        return f"<{self}>"
